import logging

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.db.models.aggregates import Count
from django.utils.functional import classproperty
from django.utils.translation import gettext as _

import reversion
from reversion import set_comment
from templated_email import send_templated_mail

from .util.celery_progress import ProgressRecorder, recorded_task
from .util.core_helpers import get_site_preferences


class SolveOption:
    """Define a solve option for one or more data checks.

    Solve options are used in order to give the data admin typical
    solutions to a data issue detected by a data check.

    Example definition

    .. code-block:: python

        from aleksis.core.data_checks import SolveOption
        from django.utils.translation import gettext as _

        class DeleteSolveOption(SolveOption):
            name = "delete" # has to be unqiue
            verbose_name = _("Delete") # should make use of i18n

            @classmethod
            def solve(cls, check_result: "DataCheckResult"):
                check_result.related_object.delete()
                check_result.delete()

    After the solve option has been successfully executed,
    the corresponding data check result has to be deleted.
    """

    name: str = "default"
    verbose_name: str = ""

    @classmethod
    def solve(cls, check_result: "DataCheckResult"):
        pass


class IgnoreSolveOption(SolveOption):
    """Mark the object with data issues as solved."""

    name = "ignore"
    verbose_name = _("Ignore problem")

    @classmethod
    def solve(cls, check_result: "DataCheckResult"):
        """Mark the object as solved without doing anything more."""
        check_result.solved = True
        check_result.save()


class DataCheck:
    """Define a data check.

    Data checks should be used to search objects of
    any type which are broken or need some extra action.

    Defining data checks
    --------------------
    Data checks are defined by inheriting from the class DataCheck
    and registering the inherited class in the data check registry.

    Example:

    ``data_checks.py``
    ******************

    .. code-block:: python

        from aleksis.core.data_checks import DataCheck, DATA_CHECK_REGISTRY
        from django.utils.translation import gettext as _

        class ExampleDataCheck(DataCheck):
            name = "example" # has to be unique
            verbose_name = _("Ensure that there are no examples.")
            problem_name = _("There is an example.") # should both make use of i18n

            solve_options = {
                IgnoreSolveOption.name: IgnoreSolveOption
            }

            @classmethod
            def check_data(cls):
                from example_app.models import ExampleModel

                wrong_examples = ExampleModel.objects.filter(wrong_value=True)

                for example in wrong_examples:
                    cls.register_result(example)

    ``models.py``
    *************

    .. code-block:: python

        from .data_checks import ExampleDataCheck

        # ...

        class ExampleModel(Model):
            data_checks = [ExampleDataCheck]


    Solve options are used in order to give the data admin typical solutions to this specific issue.
    They are defined by inheriting from SolveOption.
    More information about defining solve options can be find there.

    The dictionary ``solve_options`` should include at least the IgnoreSolveOption,
    but preferably also own solve options. The keys in this dictionary
    have to be ``<YourOption>SolveOption.name``
    and the values must be the corresponding solve option classes.

    The class method ``check_data`` does the actual work. In this method
    your code should find all objects with issues and should register
    them in the result database using the class method ``register_result``.

    Data checks have to be registered in their corresponding model.
    This can be done by adding a list ``data_checks``
    containing the data check classes.

    Executing data checks
    ---------------------
    The data checks can be executed by using the
    celery task named ``aleksis.core.data_checks.check_data``.
    We recommend to create a periodic task in the backend
    which executes ``check_data`` on a regular base (e. g. every day).

    .. warning::
        To use the option described above, you must have setup celery properly.

    Notifications about results
    ---------------------------
    The data check tasks can notify persons via email
    if there are new data issues. You can set these persons
    by adding them to the preference
    ``Email recipients for data checks problem emails`` in the site configuration.

    To enable this feature, you also have to activate
    the preference ``Send emails if data checks detect problems``.
    """  # noqa: D412

    name: str = ""
    verbose_name: str = ""
    problem_name: str = ""

    solve_options = {IgnoreSolveOption.name: IgnoreSolveOption}

    _current_results = []

    @classmethod
    def check_data(cls):
        """Find all objects with data issues and register them."""
        pass

    @classmethod
    def run_check_data(cls):
        """Wrap ``check_data`` to ensure that post-processing tasks are run."""
        cls.check_data()
        cls.delete_old_results()

    @classmethod
    def solve(cls, check_result: "DataCheckResult", solve_option: str):
        """Execute a solve option for an object detected by this check.

        :param check_result: The result item from database
        :param solve_option: The name of the solve option that should be executed
        """
        with reversion.create_revision():
            solve_option_obj = cls.solve_options[solve_option]
            set_comment(
                _(
                    f"Solve option '{solve_option_obj.verbose_name}' "
                    f"for data check '{cls.verbose_name}'"
                )
            )
            solve_option_obj.solve(check_result)

    @classmethod
    def register_result(cls, instance) -> "DataCheckResult":
        """Register an object with data issues in the result database.

        :param instance: The affected object
        :return: The database entry
        """
        from aleksis.core.models import DataCheckResult

        ct = ContentType.objects.get_for_model(instance)
        result, __ = DataCheckResult.objects.get_or_create(
            check=cls.name, content_type=ct, object_id=instance.id
        )

        # Track all existing problems (for deleting old results)
        cls._current_results.append(result)

        return result

    @classmethod
    def delete_old_results(cls):
        """Delete old data check results for problems which exist no longer."""
        DataCheckResult = apps.get_model("core", "DataCheckResult")

        pks = [r.pk for r in cls._current_results]
        old_results = DataCheckResult.objects.filter(check=cls.name).exclude(pk__in=pks)

        if old_results:
            logging.info(f"Delete {old_results.count()} old data check results.")
            old_results.delete()

        # Reset list with existing problems
        cls._current_results = []


class DataCheckRegistry:
    """Create central registry for all data checks in AlekSIS."""

    data_checks: set = set()

    @classproperty
    def data_checks_by_name(cls):
        return {check.name: check for check in cls.data_checks}

    @classproperty
    def data_checks_choices(cls):
        return [(check.name, check.verbose_name) for check in cls.data_checks]


@recorded_task
def check_data(recorder: ProgressRecorder):
    """Execute all registered data checks and send email if activated."""
    for check in recorder.iterate(DataCheckRegistry.data_checks):
        logging.info(f"Run check: {check.verbose_name}")
        check.run_check_data()

    if get_site_preferences()["general__data_checks_send_emails"]:
        send_emails_for_data_checks()


def send_emails_for_data_checks():
    """Notify one or more recipients about new problems with data.

    Recipients can be set in dynamic preferences.
    """
    from .models import DataCheckResult  # noqa

    results = DataCheckResult.objects.filter(solved=False, sent=False)

    if results.exists():
        results_by_check = results.values("check").annotate(count=Count("check"))

        results_with_checks = []
        for result in results_by_check:
            results_with_checks.append(
                (DataCheckRegistry.data_checks_by_name[result["check"]], result["count"])
            )

        recipient_list = [
            p.mail_sender
            for p in get_site_preferences()["general__data_checks_recipients"]
            if p.email
        ]

        for group in get_site_preferences()["general__data_checks_recipient_groups"]:
            recipient_list += [p.mail_sender for p in group.announcement_recipients if p.email]

        send_templated_mail(
            template_name="data_checks",
            from_email=get_site_preferences()["mail__address"],
            recipient_list=recipient_list,
            context={"results": results_with_checks},
        )

        logging.info("Sent notification email because of unsent data checks")

        results.update(sent=True)


class DeactivateDashboardWidgetSolveOption(SolveOption):
    name = "deactivate_dashboard_widget"
    verbose_name = _("Deactivate DashboardWidget")

    @classmethod
    def solve(cls, check_result: "DataCheckResult"):
        widget = check_result.related_object
        widget.active = False
        widget.save()
        check_result.delete()


class BrokenDashboardWidgetDataCheck(DataCheck):
    name = "broken_dashboard_widgets"
    verbose_name = _("Ensure that there are no broken DashboardWidgets.")
    problem_name = _("The DashboardWidget was reported broken automatically.")
    solve_options = {
        IgnoreSolveOption.name: IgnoreSolveOption,
        DeactivateDashboardWidgetSolveOption.name: DeactivateDashboardWidgetSolveOption,
    }

    @classmethod
    def check_data(cls):
        from .models import DashboardWidget

        broken_widgets = DashboardWidget.objects.filter(broken=True, active=True)

        for widget in broken_widgets:
            logging.info("Check DashboardWidget %s", widget)
            cls.register_result(widget)



import rules

from .models import AdditionalField, Announcement, Group, GroupType, Person
from .util.predicates import (
    contains_site_preference_value,
    has_any_object,
    has_global_perm,
    has_object_perm,
    has_person,
    is_current_person,
    is_group_owner,
    is_notification_recipient,
    is_site_preference_set,
)

rules.add_perm("core", rules.always_allow)

# View dashboard
view_dashboard_predicate = is_site_preference_set("general", "anonymous_dashboard") | has_person
rules.add_perm("core.view_dashboard_rule", view_dashboard_predicate)

# View notifications
rules.add_perm("core.view_notifications_rule", has_person)

# Use search
search_predicate = has_person & has_global_perm("core.search")
rules.add_perm("core.search_rule", search_predicate)

# View persons
view_persons_predicate = has_person & (
    has_global_perm("core.view_person") | has_any_object("core.view_person", Person)
)
rules.add_perm("core.view_persons_rule", view_persons_predicate)

# View person
view_person_predicate = has_person & (
    has_global_perm("core.view_person") | has_object_perm("core.view_person") | is_current_person
)
rules.add_perm("core.view_person_rule", view_person_predicate)

# View person address
view_address_predicate = has_person & (
    has_global_perm("core.view_address") | has_object_perm("core.view_address") | is_current_person
)
rules.add_perm("core.view_address_rule", view_address_predicate)

# View person contact details
view_contact_details_predicate = has_person & (
    has_global_perm("core.view_contact_details")
    | has_object_perm("core.view_contact_details")
    | is_current_person
)
rules.add_perm("core.view_contact_details_rule", view_contact_details_predicate)

# View person photo
view_photo_predicate = has_person & (
    has_global_perm("core.view_photo") | has_object_perm("core.view_photo") | is_current_person
)
rules.add_perm("core.view_photo_rule", view_photo_predicate)

# View persons groups
view_groups_predicate = has_person & (
    has_global_perm("core.view_person_groups")
    | has_object_perm("core.view_person_groups")
    | is_current_person
)
rules.add_perm("core.view_person_groups_rule", view_groups_predicate)

# Edit person
edit_person_predicate = has_person & (
    has_global_perm("core.change_person")
    | has_object_perm("core.change_person")
    | is_current_person & is_site_preference_set("account", "editable_fields_person")
)
rules.add_perm("core.edit_person_rule", edit_person_predicate)

# Delete person
delete_person_predicate = has_person & (
    has_global_perm("core.delete_person") | has_object_perm("core.delete_person")
)
rules.add_perm("core.delete_person_rule", delete_person_predicate)

# Link persons with accounts
link_persons_accounts_predicate = has_person & has_global_perm("core.link_persons_accounts")
rules.add_perm("core.link_persons_accounts_rule", link_persons_accounts_predicate)

# View groups
view_groups_predicate = has_person & (
    has_global_perm("core.view_group") | has_any_object("core.view_group", Group)
)
rules.add_perm("core.view_groups_rule", view_groups_predicate)

# View group
view_group_predicate = has_person & (
    has_global_perm("core.view_group") | has_object_perm("core.view_group")
)
rules.add_perm("core.view_group_rule", view_group_predicate)

# Edit group
edit_group_predicate = has_person & (
    has_global_perm("core.change_group") | has_object_perm("core.change_group")
)
rules.add_perm("core.edit_group_rule", edit_group_predicate)

# Delete group
delete_group_predicate = has_person & (
    has_global_perm("core.delete_group") | has_object_perm("core.delete_group")
)
rules.add_perm("core.delete_group_rule", delete_group_predicate)

# Assign child groups to groups
assign_child_groups_to_groups_predicate = has_person & has_global_perm(
    "core.assign_child_groups_to_groups"
)
rules.add_perm("core.assign_child_groups_to_groups_rule", assign_child_groups_to_groups_predicate)

# Edit school information
edit_school_information_predicate = has_person & has_global_perm("core.change_school")
rules.add_perm("core.edit_school_information_rule", edit_school_information_predicate)

# Manage data
manage_data_predicate = has_person & has_global_perm("core.manage_data")
rules.add_perm("core.manage_data_rule", manage_data_predicate)

# Mark notification as read
mark_notification_as_read_predicate = has_person & is_notification_recipient
rules.add_perm("core.mark_notification_as_read_rule", mark_notification_as_read_predicate)

# View announcements
view_announcements_predicate = has_person & (
    has_global_perm("core.view_announcement")
    | has_any_object("core.view_announcement", Announcement)
)
rules.add_perm("core.view_announcements_rule", view_announcements_predicate)

# Create or edit announcement
create_or_edit_announcement_predicate = has_person & (
    has_global_perm("core.add_announcement")
    & (has_global_perm("core.change_announcement") | has_object_perm("core.change_announcement"))
)
rules.add_perm("core.create_or_edit_announcement_rule", create_or_edit_announcement_predicate)

# Delete announcement
delete_announcement_predicate = has_person & (
    has_global_perm("core.delete_announcement") | has_object_perm("core.delete_announcement")
)
rules.add_perm("core.delete_announcement_rule", delete_announcement_predicate)

# Use impersonate
impersonate_predicate = has_person & has_global_perm("core.impersonate")
rules.add_perm("core.impersonate_rule", impersonate_predicate)

# View system status
view_system_status_predicate = has_person & has_global_perm("core.view_system_status")
rules.add_perm("core.view_system_status_rule", view_system_status_predicate)

# View people menu (persons + objects)
rules.add_perm(
    "core.view_people_menu_rule",
    has_person
    & (
        view_persons_predicate
        | view_groups_predicate
        | link_persons_accounts_predicate
        | assign_child_groups_to_groups_predicate
    ),
)

# View person personal details
view_personal_details_predicate = has_person & (
    has_global_perm("core.view_personal_details")
    | has_object_perm("core.view_personal_details")
    | is_current_person
)
rules.add_perm("core.view_personal_details_rule", view_personal_details_predicate)

# Change site preferences
change_site_preferences = has_person & (
    has_global_perm("core.change_site_preferences")
    | has_object_perm("core.change_site_preferences")
)
rules.add_perm("core.change_site_preferences_rule", change_site_preferences)

# Change person preferences
change_person_preferences = has_person & (
    has_global_perm("core.change_person_preferences")
    | has_object_perm("core.change_person_preferences")
    | is_current_person
)
rules.add_perm("core.change_person_preferences_rule", change_person_preferences)

# Change group preferences
change_group_preferences = has_person & (
    has_global_perm("core.change_group_preferences")
    | has_object_perm("core.change_group_preferences")
    | is_group_owner
)
rules.add_perm("core.change_group_preferences_rule", change_group_preferences)


# Edit additional field
change_additional_field_predicate = has_person & (
    has_global_perm("core.change_additionalfield") | has_object_perm("core.change_additionalfield")
)
rules.add_perm("core.change_additionalfield_rule", change_additional_field_predicate)

# Edit additional field
create_additional_field_predicate = has_person & (
    has_global_perm("core.create_additionalfield") | has_object_perm("core.create_additionalfield")
)
rules.add_perm("core.create_additionalfield_rule", create_additional_field_predicate)


# Delete additional field
delete_additional_field_predicate = has_person & (
    has_global_perm("core.delete_additionalfield") | has_object_perm("core.delete_additionalfield")
)
rules.add_perm("core.delete_additionalfield_rule", delete_additional_field_predicate)

# View additional fields
view_additional_fields_predicate = has_person & (
    has_global_perm("core.view_additionalfield")
    | has_any_object("core.view_additionalfield", AdditionalField)
)
rules.add_perm("core.view_additionalfields_rule", view_additional_fields_predicate)

# Edit group type
change_group_type_predicate = has_person & (
    has_global_perm("core.change_grouptype") | has_object_perm("core.change_grouptype")
)
rules.add_perm("core.edit_grouptype_rule", change_group_type_predicate)

# Create group type
create_group_type_predicate = has_person & (
    has_global_perm("core.create_grouptype") | has_object_perm("core.change_grouptype")
)
rules.add_perm("core.create_grouptype_rule", create_group_type_predicate)


# Delete group type
delete_group_type_predicate = has_person & (
    has_global_perm("core.delete_grouptype") | has_object_perm("core.delete_grouptype")
)
rules.add_perm("core.delete_grouptype_rule", delete_group_type_predicate)

# View group types
view_group_types_predicate = has_person & (
    has_global_perm("core.view_grouptype") | has_any_object("core.view_grouptype", GroupType)
)
rules.add_perm("core.view_grouptypes_rule", view_group_types_predicate)

# Create person
create_person_predicate = has_person & (
    has_global_perm("core.create_person") | has_object_perm("core.create_person")
)
rules.add_perm("core.create_person_rule", create_person_predicate)

# Create group
create_group_predicate = has_person & (
    has_global_perm("core.create_group") | has_object_perm("core.create_group")
)
rules.add_perm("core.create_group_rule", create_group_predicate)

# School years
view_school_term_predicate = has_person & has_global_perm("core.view_schoolterm")
rules.add_perm("core.view_schoolterm_rule", view_school_term_predicate)

create_school_term_predicate = has_person & has_global_perm("core.add_schoolterm")
rules.add_perm("core.create_schoolterm_rule", create_school_term_predicate)

edit_school_term_predicate = has_person & has_global_perm("core.change_schoolterm")
rules.add_perm("core.edit_schoolterm_rule", edit_school_term_predicate)

# View admin menu
view_admin_menu_predicate = has_person & (
    manage_data_predicate
    | view_school_term_predicate
    | impersonate_predicate
    | view_system_status_predicate
    | view_announcements_predicate
)
rules.add_perm("core.view_admin_menu_rule", view_admin_menu_predicate)

# View group stats
view_group_stats_predicate = has_person & (
    has_global_perm("core.view_group_stats") | has_object_perm("core.view_group_stats")
)
rules.add_perm("core.view_group_stats_rule", view_group_stats_predicate)

# View data check results
view_data_check_results_predicate = has_person & has_global_perm("core.view_datacheckresult")
rules.add_perm("core.view_datacheckresults_rule", view_data_check_results_predicate)

# Run data checks
run_data_checks_predicate = (
    has_person & view_data_check_results_predicate & has_global_perm("core.run_data_checks")
)
rules.add_perm("core.run_data_checks_rule", run_data_checks_predicate)

# Solve data problems
solve_data_problem_predicate = (
    has_person & view_data_check_results_predicate & has_global_perm("core.solve_data_problem")
)
rules.add_perm("core.solve_data_problem_rule", solve_data_problem_predicate)

view_dashboard_widget_predicate = has_person & has_global_perm("core.view_dashboardwidget")
rules.add_perm("core.view_dashboardwidget_rule", view_dashboard_widget_predicate)

create_dashboard_widget_predicate = has_person & has_global_perm("core.add_dashboardwidget")
rules.add_perm("core.create_dashboardwidget_rule", create_dashboard_widget_predicate)

edit_dashboard_widget_predicate = has_person & has_global_perm("core.change_dashboardwidget")
rules.add_perm("core.edit_dashboardwidget_rule", edit_dashboard_widget_predicate)

delete_dashboard_widget_predicate = has_person & has_global_perm("core.delete_dashboardwidget")
rules.add_perm("core.delete_dashboardwidget_rule", delete_dashboard_widget_predicate)

edit_dashboard_predicate = is_site_preference_set("general", "dashboard_editing") & has_person
rules.add_perm("core.edit_dashboard_rule", edit_dashboard_predicate)

edit_default_dashboard_predicate = has_person & has_global_perm("core.edit_default_dashboard")
rules.add_perm("core.edit_default_dashboard_rule", edit_default_dashboard_predicate)

# django-allauth
can_register_predicate = is_site_preference_set(section="auth", pref="signup_enabled")
rules.add_perm("core.can_register", can_register_predicate)

can_change_password_predicate = is_site_preference_set(section="auth", pref="allow_password_change")
rules.add_perm("core.can_change_password", can_change_password_predicate)

# OAuth2 permissions
add_oauth_applications_predicate = has_person & has_global_perm("core.add_oauth_applications")
rules.add_perm("core.add_oauth_applications_rule", add_oauth_applications_predicate)

list_oauth_applications_predicate = has_person & has_global_perm("core.list_oauth_applications")
rules.add_perm("core.list_oauth_applications_rule", list_oauth_applications_predicate)

view_oauth_applications_predicate = has_person & has_global_perm("core.view_oauth_applications")
rules.add_perm("core.view_oauth_applications_rule", view_oauth_applications_predicate)

update_oauth_applications_predicate = has_person & has_global_perm("core.update_oauth_applications")
rules.add_perm("core.update_oauth_applications_rule", update_oauth_applications_predicate)

delete_oauth_applications_predicate = has_person & has_global_perm("core.delete_oauth_applications")
rules.add_perm("core.delete_oauth_applications_rule", delete_oauth_applications_predicate)

# Upload and browse files via CKEditor
upload_files_ckeditor_predicate = has_person & has_global_perm("core.upload_files_ckeditor")
rules.add_perm("core.upload_files_ckeditor_rule", upload_files_ckeditor_predicate)

test_pdf_generation_predicate = has_person & has_global_perm("core.test_pdf")
rules.add_perm("core.test_pdf_rule", test_pdf_generation_predicate)

# Generate rules for syncable fields
for field in Person._meta.fields:
    perm = (
        has_global_perm("core.edit_person")
        | has_object_perm("core.edit_person")
        | (
            is_current_person
            & contains_site_preference_value("account", "editable_fields_person", field.name)
        )
    )
    rules.add_perm(f"core.change_person_field_{field.name}_rule", perm)



from hattori.base import BaseAnonymizer, faker

from .models import Person


class PersonAnonymizer(BaseAnonymizer):
    model = Person

    attributes = [
        ("first_name", faker.first_name),
        ("last_name", faker.last_name),
        ("additional_name", ""),
        ("short_name", lambda **kwargs: faker.pystr(min_chars=3, max_chars=5, **kwargs)),
        ("street", faker.street_name),
        ("housenumber", faker.building_number),
        ("postal_code", faker.postcode),
        ("place", faker.city),
        ("phone_number", ""),
        ("mobile_number", ""),
        ("email", faker.email),
        (
            "date_of_birth",
            lambda **kwargs: faker.date_of_birth(minimum_age=8, maximum_age=66, **kwargs),
        ),
        ("photo", ""),
    ]



from typing import Any, Optional

import django.apps
from django.apps import apps
from django.http import HttpRequest
from django.utils.module_loading import autodiscover_modules

from dynamic_preferences.registries import preference_models
from health_check.plugins import plugin_dir

from .registries import (
    group_preferences_registry,
    person_preferences_registry,
    site_preferences_registry,
)
from .util.apps import AppConfig
from .util.core_helpers import get_or_create_favicon, has_person
from .util.sass_helpers import clean_scss


class CoreConfig(AppConfig):
    name = "aleksis.core"
    verbose_name = "AlekSIS — The Free School Information System"
    dist_name = "AlekSIS-Core"

    urls = {
        "Repository": "https://edugit.org/AlekSIS/official/AlekSIS/",
    }
    licence = "EUPL-1.2+"
    copyright_info = (
        ([2017, 2018, 2019, 2020, 2021], "Jonathan Weth", "wethjo@katharineum.de"),
        ([2017, 2018, 2019, 2020], "Frank Poetzsch-Heffter", "p-h@katharineum.de"),
        ([2018, 2019, 2020, 2021], "Julian Leucker", "leuckeju@katharineum.de"),
        ([2018, 2019, 2020, 2021], "Hangzhi Yu", "yuha@katharineum.de"),
        ([2019, 2020, 2021], "Dominik George", "dominik.george@teckids.org"),
        ([2019, 2020, 2021], "Tom Teichler", "tom.teichler@teckids.org"),
        ([2019], "mirabilos", "thorsten.glaser@teckids.org"),
        ([2021], "Lloyd Meins", "meinsll@katharineum.de"),
        ([2021], "magicfelix", "felix@felix-zauberer.de"),
    )

    def ready(self):
        super().ready()

        from django.conf import settings  # noqa

        # Autodiscover various modules defined by AlekSIS
        autodiscover_modules("form_extensions", "model_extensions", "checks")

        sitepreferencemodel = self.get_model("SitePreferenceModel")
        personpreferencemodel = self.get_model("PersonPreferenceModel")
        grouppreferencemodel = self.get_model("GroupPreferenceModel")

        preference_models.register(sitepreferencemodel, site_preferences_registry)
        preference_models.register(personpreferencemodel, person_preferences_registry)
        preference_models.register(grouppreferencemodel, group_preferences_registry)

        self._load_data_checks()

        from .health_checks import (
            BackupJobHealthCheck,
            DataChecksHealthCheckBackend,
            DbBackupAgeHealthCheck,
            MediaBackupAgeHealthCheck,
        )

        plugin_dir.register(DataChecksHealthCheckBackend)
        plugin_dir.register(DbBackupAgeHealthCheck)
        plugin_dir.register(MediaBackupAgeHealthCheck)
        plugin_dir.register(BackupJobHealthCheck)

    @classmethod
    def _load_data_checks(cls):
        """Get all data checks from all loaded models."""
        from aleksis.core.data_checks import DataCheckRegistry

        data_checks = set()
        for model in apps.get_models():
            data_checks.update(getattr(model, "data_checks", []))
        DataCheckRegistry.data_checks = data_checks

    def preference_updated(
        self,
        sender: Any,
        section: Optional[str] = None,
        name: Optional[str] = None,
        old_value: Optional[Any] = None,
        new_value: Optional[Any] = None,
        **kwargs,
    ) -> None:
        from django.conf import settings  # noqa

        if section == "theme":
            if name in ("primary", "secondary"):
                clean_scss()
            elif name in ("favicon", "pwa_icon"):
                from favicon.models import Favicon  # noqa

                is_favicon = name == "favicon"

                if new_value:
                    Favicon.on_site.update_or_create(
                        title=name, defaults={"isFavicon": is_favicon, "faviconImage": new_value},
                    )
                else:
                    Favicon.on_site.filter(title=name, isFavicon=is_favicon).delete()
                    if name in settings.DEFAULT_FAVICON_PATHS:
                        get_or_create_favicon(
                            name, settings.DEFAULT_FAVICON_PATHS[name], is_favicon=is_favicon
                        )

    def post_migrate(
        self,
        app_config: django.apps.AppConfig,
        verbosity: int,
        interactive: bool,
        using: str,
        **kwargs,
    ) -> None:
        from django.conf import settings  # noqa

        super().post_migrate(app_config, verbosity, interactive, using, **kwargs)

        # Ensure presence of an OTP YubiKey default config
        apps.get_model("otp_yubikey", "ValidationService").objects.using(using).update_or_create(
            name="default", defaults={"use_ssl": True, "param_sl": "", "param_timeout": ""}
        )

        # Ensure that default Favicon object exists
        for name, default in settings.DEFAULT_FAVICON_PATHS.items():
            get_or_create_favicon(name, default, is_favicon=name == "favicon")

    def user_logged_in(
        self, sender: type, request: Optional[HttpRequest], user: "User", **kwargs
    ) -> None:
        if has_person(user):
            # Save the associated person to pick up defaults
            user.person.save()



from celery_haystack.indexes import CelerySearchIndex as BaseSearchIndex
from haystack import indexes

# Not used here, but simplifies imports for apps
Indexable = indexes.Indexable  # noqa


class SearchIndex(BaseSearchIndex):
    """Base class for search indexes on AlekSIS models.

    It provides a default document field caleld text and exects
    the related model in the model attribute.
    """

    text = indexes.EdgeNgramField(document=True, use_template=True)

    def get_model(self):
        return self.model



from datetime import datetime, time
from typing import Callable, Sequence

from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db.models import QuerySet
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _

from allauth.account.adapter import get_adapter
from allauth.account.forms import SignupForm
from allauth.account.utils import get_user_model, setup_user_email
from django_select2.forms import ModelSelect2MultipleWidget, ModelSelect2Widget, Select2Widget
from dynamic_preferences.forms import PreferenceForm
from guardian.core import ObjectPermissionChecker
from material import Fieldset, Layout, Row

from .mixins import ExtensibleForm, SchoolTermRelatedExtensibleForm
from .models import (
    AdditionalField,
    Announcement,
    DashboardWidget,
    Group,
    GroupType,
    Person,
    SchoolTerm,
)
from .registries import (
    group_preferences_registry,
    person_preferences_registry,
    site_preferences_registry,
)
from .util.core_helpers import get_site_preferences


class PersonAccountForm(forms.ModelForm):
    """Form to assign user accounts to persons in the frontend."""

    class Meta:
        model = Person
        fields = ["last_name", "first_name", "user"]
        widgets = {"user": Select2Widget(attrs={"class": "browser-default"})}

    new_user = forms.CharField(required=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Fields displayed only for informational purposes
        self.fields["first_name"].disabled = True
        self.fields["last_name"].disabled = True

    def clean(self) -> None:
        user = get_user_model()

        if self.cleaned_data.get("new_user", None):
            if self.cleaned_data.get("user", None):
                # The user selected both an existing user and provided a name to create a new one
                self.add_error(
                    "new_user",
                    _("You cannot set a new username when also selecting an existing user."),
                )
            elif user.objects.filter(username=self.cleaned_data["new_user"]).exists():
                # The user tried to create a new user with the name of an existing user
                self.add_error("new_user", _("This username is already in use."))
            else:
                # Create new User object and assign to form field for existing user
                new_user_obj = user.objects.create_user(
                    self.cleaned_data["new_user"],
                    self.instance.email,
                    first_name=self.instance.first_name,
                    last_name=self.instance.last_name,
                )

                self.cleaned_data["user"] = new_user_obj


# Formset for batch-processing of assignments of users to persons
PersonsAccountsFormSet = forms.modelformset_factory(
    Person, form=PersonAccountForm, max_num=0, extra=0
)


class EditPersonForm(ExtensibleForm):
    """Form to edit an existing person object in the frontend."""

    layout = Layout(
        Fieldset(
            _("Base data"),
            "short_name",
            Row("user", "primary_group"),
            "is_active",
            Row("first_name", "additional_name", "last_name"),
        ),
        Fieldset(_("Address"), Row("street", "housenumber"), Row("postal_code", "place")),
        Fieldset(_("Contact data"), "email", Row("phone_number", "mobile_number")),
        Fieldset(
            _("Advanced personal data"), Row("sex", "date_of_birth"), Row("photo"), "guardians",
        ),
    )

    class Meta:
        model = Person
        fields = [
            "user",
            "is_active",
            "first_name",
            "last_name",
            "additional_name",
            "short_name",
            "street",
            "housenumber",
            "postal_code",
            "place",
            "phone_number",
            "mobile_number",
            "email",
            "date_of_birth",
            "sex",
            "photo",
            "guardians",
            "primary_group",
        ]
        widgets = {
            "user": Select2Widget(attrs={"class": "browser-default"}),
            "primary_group": ModelSelect2Widget(
                search_fields=["name__icontains", "short_name__icontains"],
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            ),
            "guardians": ModelSelect2MultipleWidget(
                search_fields=[
                    "first_name__icontains",
                    "last_name__icontains",
                    "short_name__icontains",
                ],
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            ),
        }

    new_user = forms.CharField(
        required=False, label=_("New user"), help_text=_("Create a new account")
    )

    def __init__(self, request: HttpRequest, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Disable non-editable fields
        person_fields = set([field.name for field in Person.syncable_fields()]).intersection(
            set(self.fields)
        )

        if self.instance:
            checker = ObjectPermissionChecker(request.user)
            checker.prefetch_perms([self.instance])

            for field in person_fields:
                if not checker.has_perm(f"core.change_person_field_{field}", self.instance):
                    self.fields[field].disabled = True

    def clean(self) -> None:
        # Use code implemented in dedicated form to verify user selection
        return PersonAccountForm.clean(self)


class EditGroupForm(SchoolTermRelatedExtensibleForm):
    """Form to edit an existing group in the frontend."""

    layout = Layout(
        Fieldset(_("School term"), "school_term"),
        Fieldset(_("Common data"), "name", "short_name", "group_type"),
        Fieldset(_("Persons"), "members", "owners", "parent_groups"),
        Fieldset(_("Additional data"), "additional_fields"),
    )

    class Meta:
        model = Group
        exclude = []
        widgets = {
            "members": ModelSelect2MultipleWidget(
                search_fields=[
                    "first_name__icontains",
                    "last_name__icontains",
                    "short_name__icontains",
                ],
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            ),
            "owners": ModelSelect2MultipleWidget(
                search_fields=[
                    "first_name__icontains",
                    "last_name__icontains",
                    "short_name__icontains",
                ],
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            ),
            "parent_groups": ModelSelect2MultipleWidget(
                search_fields=["name__icontains", "short_name__icontains"],
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            ),
            "additional_fields": ModelSelect2MultipleWidget(
                search_fields=["title__icontains",],
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            ),
        }


class AnnouncementForm(ExtensibleForm):
    """Form to create or edit an announcement in the frontend."""

    valid_from = forms.DateTimeField(required=False)
    valid_until = forms.DateTimeField(required=False)

    valid_from_date = forms.DateField(label=_("Date"))
    valid_from_time = forms.TimeField(label=_("Time"))

    valid_until_date = forms.DateField(label=_("Date"))
    valid_until_time = forms.TimeField(label=_("Time"))

    persons = forms.ModelMultipleChoiceField(
        queryset=Person.objects.all(),
        label=_("Persons"),
        required=False,
        widget=ModelSelect2MultipleWidget(
            search_fields=[
                "first_name__icontains",
                "last_name__icontains",
                "short_name__icontains",
            ],
            attrs={"data-minimum-input-length": 0, "class": "browser-default"},
        ),
    )
    groups = forms.ModelMultipleChoiceField(
        queryset=None,
        label=_("Groups"),
        required=False,
        widget=ModelSelect2MultipleWidget(
            search_fields=["name__icontains", "short_name__icontains",],
            attrs={"data-minimum-input-length": 0, "class": "browser-default"},
        ),
    )

    layout = Layout(
        Fieldset(
            _("From when until when should the announcement be displayed?"),
            Row("valid_from_date", "valid_from_time", "valid_until_date", "valid_until_time"),
        ),
        Fieldset(_("Who should see the announcement?"), Row("groups", "persons")),
        Fieldset(_("Write your announcement:"), "title", "description"),
    )

    def __init__(self, *args, **kwargs):
        if "instance" not in kwargs:
            # Default to today and whole day for new announcements
            kwargs["initial"] = {
                "valid_from_date": datetime.now(),
                "valid_from_time": time(0, 0),
                "valid_until_date": datetime.now(),
                "valid_until_time": time(23, 59),
            }
        else:
            announcement = kwargs["instance"]

            # Fill special fields from given announcement instance
            kwargs["initial"] = {
                "valid_from_date": announcement.valid_from.date(),
                "valid_from_time": announcement.valid_from.time(),
                "valid_until_date": announcement.valid_until.date(),
                "valid_until_time": announcement.valid_until.time(),
                "groups": announcement.get_recipients_for_model(Group),
                "persons": announcement.get_recipients_for_model(Person),
            }

        super().__init__(*args, **kwargs)

        self.fields["groups"].queryset = Group.objects.for_current_school_term_or_all()

    def clean(self):
        data = super().clean()

        # Combine date and time fields into datetime objects
        valid_from = datetime.combine(data["valid_from_date"], data["valid_from_time"])
        valid_until = datetime.combine(data["valid_until_date"], data["valid_until_time"])

        # Sanity check validity range
        if valid_until < datetime.now():
            raise ValidationError(
                _("You are not allowed to create announcements which are only valid in the past.")
            )
        elif valid_from > valid_until:
            raise ValidationError(
                _("The from date and time must be earlier then the until date and time.")
            )

        # Inject real time data if all went well
        data["valid_from"] = valid_from
        data["valid_until"] = valid_until

        # Ensure at least one group or one person is set as recipient
        if "groups" not in data and "persons" not in data:
            raise ValidationError(_("You need at least one recipient."))

        # Unwrap all recipients into single user objects and generate final list
        data["recipients"] = []
        data["recipients"] += data.get("groups", [])
        data["recipients"] += data.get("persons", [])

        return data

    def save(self, _=False):
        # Save announcement, respecting data injected in clean()
        if self.instance is None:
            self.instance = Announcement()
        self.instance.valid_from = self.cleaned_data["valid_from"]
        self.instance.valid_until = self.cleaned_data["valid_until"]
        self.instance.title = self.cleaned_data["title"]
        self.instance.description = self.cleaned_data["description"]
        self.instance.save()

        # Save recipients
        self.instance.recipients.all().delete()
        for recipient in self.cleaned_data["recipients"]:
            self.instance.recipients.create(recipient=recipient)
        self.instance.save()

        return self.instance

    class Meta:
        model = Announcement
        exclude = []


class ChildGroupsForm(forms.Form):
    """Inline form for group editing to select child groups."""

    child_groups = forms.ModelMultipleChoiceField(queryset=Group.objects.all())


class SitePreferenceForm(PreferenceForm):
    """Form to edit site preferences."""

    registry = site_preferences_registry


class PersonPreferenceForm(PreferenceForm):
    """Form to edit preferences valid for one person."""

    registry = person_preferences_registry


class GroupPreferenceForm(PreferenceForm):
    """Form to edit preferences valid for members of a group."""

    registry = group_preferences_registry


class EditAdditionalFieldForm(forms.ModelForm):
    """Form to manage additional fields."""

    class Meta:
        model = AdditionalField
        exclude = []


class EditGroupTypeForm(forms.ModelForm):
    """Form to manage group types."""

    class Meta:
        model = GroupType
        exclude = []


class SchoolTermForm(ExtensibleForm):
    """Form for managing school years."""

    layout = Layout("name", Row("date_start", "date_end"))

    class Meta:
        model = SchoolTerm
        exclude = []


class DashboardWidgetOrderForm(ExtensibleForm):
    pk = forms.ModelChoiceField(
        queryset=None, widget=forms.HiddenInput(attrs={"class": "pk-input"}),
    )
    order = forms.IntegerField(initial=0, widget=forms.HiddenInput(attrs={"class": "order-input"}))

    class Meta:
        model = DashboardWidget
        fields = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Set queryset here to prevent problems with not migrated database due to special queryset
        self.fields["pk"].queryset = DashboardWidget.objects.all()


DashboardWidgetOrderFormSet = forms.formset_factory(
    form=DashboardWidgetOrderForm, max_num=0, extra=0
)


class AccountRegisterForm(SignupForm, ExtensibleForm):
    """Form to register new user accounts."""

    class Meta:
        model = Group
        fields = []

    layout = Layout(
        Fieldset(_("Base data"), Row("first_name", "last_name"),),
        Fieldset(
            _("Account data"), "username", Row("email", "email2"), Row("password1", "password2"),
        ),
        Fieldset(_("Consents"), Row("privacy_policy"),),
    )

    def __init__(self, *args, **kwargs):
        super(AccountRegisterForm, self).__init__(*args, **kwargs)
        self.fields["password1"] = forms.CharField(label=_("Password"), widget=forms.PasswordInput)

        privacy_policy = get_site_preferences()["footer__privacy_url"]

        if settings.SIGNUP_PASSWORD_ENTER_TWICE:
            self.fields["password2"] = forms.CharField(
                label=_("Password (again)"), widget=forms.PasswordInput
            )

        self.fields["first_name"] = forms.CharField(required=True,)

        self.fields["last_name"] = forms.CharField(required=True,)

        self.fields["privacy_policy"] = forms.BooleanField(
            help_text=_(
                f"I have read the <a href='{privacy_policy}'>Privacy policy</a>"
                " and agree with them."
            ),
            required=True,
        )

    def clean(self):
        super(AccountRegisterForm, self).clean()

        dummy_user = get_user_model()
        password = self.cleaned_data.get("password1")
        if password:
            try:
                get_adapter().clean_password(password, user=dummy_user)
            except forms.ValidationError as e:
                self.add_error("password1", e)

        if (
            settings.SIGNUP_PASSWORD_ENTER_TWICE
            and "password1" in self.cleaned_data
            and "password2" in self.cleaned_data
        ):
            if self.cleaned_data["password1"] != self.cleaned_data["password2"]:
                self.add_error(
                    "password2", _("You must type the same password each time."),
                )
        return self.cleaned_data

    def save(self, request):
        adapter = get_adapter(request)
        user = adapter.new_user(request)
        adapter.save_user(request, user, self)
        Person.objects.create(
            first_name=self.cleaned_data["first_name"],
            last_name=self.cleaned_data["last_name"],
            email=self.cleaned_data["email"],
            user=user,
        )
        self.custom_signup(request, user)
        setup_user_email(request, user, [])
        return user


class ActionForm(forms.Form):
    """Generic form for executing actions on multiple items of a queryset.

    This should be used together with a ``Table`` from django-tables2
    which includes a ``SelectColumn``.

    The queryset can be defined in two different ways:
    You can use ``get_queryset`` or provide ``queryset`` as keyword argument
    at the initialization of this form class.
    If both are declared, it will use the keyword argument.

    Any actions can be defined using the ``actions`` class attribute
    or overriding the method ``get_actions``.
    The actions use the same syntax like the Django Admin actions with one important difference:
    Instead of the related model admin,
    these actions will get the related ``ActionForm`` as first argument.
    Here you can see an example for such an action:

    .. code-block:: python

        from django.utils.translation import gettext as _

        def example_action(form, request, queryset):
            # Do something with this queryset

        example_action.short_description = _("Example action")

    If you can include the ``ActionForm`` like any other form in your views,
    but you must add the request as first argument.
    When the form is valid, you should run ``execute``:

    .. code-block:: python

        from aleksis.core.forms import ActionForm

        def your_view(request, ...):
            # Something
            action_form = ActionForm(request, request.POST or None, ...)
            if request.method == "POST" and form.is_valid():
                form.execute()

            # Something
    """

    layout = Layout("action")
    actions = []

    def get_actions(self) -> Sequence[Callable]:
        """Get all defined actions."""
        return self.actions

    def _get_actions_dict(self) -> dict[str, Callable]:
        """Get all defined actions as dictionary."""
        return {value.__name__: value for value in self.get_actions()}

    def _get_action_choices(self) -> list[tuple[str, str]]:
        """Get all defined actions as Django choices."""
        return [
            (value.__name__, getattr(value, "short_description", value.__name__))
            for value in self.get_actions()
        ]

    def get_queryset(self) -> QuerySet:
        """Get the related queryset."""
        raise NotImplementedError("Queryset necessary.")

    action = forms.ChoiceField(choices=[])
    selected_objects = forms.ModelMultipleChoiceField(queryset=None)

    def __init__(self, request: HttpRequest, *args, queryset: QuerySet = None, **kwargs):
        self.request = request
        self.queryset = queryset if isinstance(queryset, QuerySet) else self.get_queryset()
        super().__init__(*args, **kwargs)
        self.fields["selected_objects"].queryset = self.queryset
        self.fields["action"].choices = self._get_action_choices()

    def execute(self) -> bool:
        """Execute the selected action on all selected objects.

        :return: If the form is not valid, it will return ``False``.
        """
        if self.is_valid():
            data = self.cleaned_data["selected_objects"]
            action = self._get_actions_dict()[self.cleaned_data["action"]]
            action(None, self.request, data)
            return True
        return False


class ListActionForm(ActionForm):
    """Generic form for executing actions on multiple items of a list of dictionaries.

    Sometimes you want to implement actions for data from different sources
    than querysets or even querysets from multiple models. For these cases,
    you can use this form.

    To provide an unique identification of each item, the dictionaries **must**
    include the attribute ``pk``. This attribute has to be unique for the whole list.
    If you don't mind this aspect, this will cause unexpected behavior.

    Any actions can be defined as described in ``ActionForm``, but, of course,
    the last argument won't be a queryset but a list of dictionaries.

    For further information on usage, you can take a look at ``ActionForm``.
    """

    selected_objects = forms.MultipleChoiceField(choices=[])

    def get_queryset(self):
        # Return None in order not to raise an unwanted exception
        return None

    def _get_dict(self) -> dict[str, dict]:
        """Get the items sorted by pk attribute."""
        return {item["pk"]: item for item in self.items}

    def _get_choices(self) -> list[tuple[str, str]]:
        """Get the items as Django choices."""
        return [(item["pk"], item["pk"]) for item in self.items]

    def _get_real_items(self, items: Sequence[dict]) -> list[dict]:
        """Get the real dictionaries from a list of pks."""
        items_dict = self._get_dict()
        real_items = []
        for item in items:
            if item not in items_dict:
                raise ValidationError(_("No valid selection."))
            real_items.append(items_dict[item])
        return real_items

    def clean_selected_objects(self) -> list[dict]:
        data = self.cleaned_data["selected_objects"]
        items = self._get_real_items(data)
        return items

    def __init__(self, request: HttpRequest, items, *args, **kwargs):
        self.items = items
        super().__init__(request, *args, **kwargs)
        self.fields["selected_objects"].choices = self._get_choices()



import pytest

from aleksis.core.models import Person

pytestmark = pytest.mark.django_db


def test_full_name():
    _person = Person.objects.create(first_name="Jane", last_name="Doe")

    assert _person.full_name == "Doe, Jane"



from django.utils.translation import gettext_lazy as _

import django_tables2 as tables
from django_tables2.utils import A


class SchoolTermTable(tables.Table):
    """Table to list persons."""

    class Meta:
        attrs = {"class": "responsive-table highlight"}

    name = tables.LinkColumn("edit_school_term", args=[A("id")])
    date_start = tables.Column()
    date_end = tables.Column()
    edit = tables.LinkColumn(
        "edit_school_term",
        args=[A("id")],
        text=_("Edit"),
        attrs={"a": {"class": "btn-flat waves-effect waves-orange orange-text"}},
        verbose_name=_("Actions"),
    )


class PersonsTable(tables.Table):
    """Table to list persons."""

    class Meta:
        attrs = {"class": "responsive-table highlight"}

    first_name = tables.LinkColumn("person_by_id", args=[A("id")])
    last_name = tables.LinkColumn("person_by_id", args=[A("id")])


class GroupsTable(tables.Table):
    """Table to list groups."""

    class Meta:
        attrs = {"class": "responsive-table highlight"}

    name = tables.LinkColumn("group_by_id", args=[A("id")])
    short_name = tables.LinkColumn("group_by_id", args=[A("id")])
    school_term = tables.Column()


class AdditionalFieldsTable(tables.Table):
    """Table to list group types."""

    class Meta:
        attrs = {"class": "responsive-table hightlight"}

    title = tables.LinkColumn("edit_additional_field_by_id", args=[A("id")])
    delete = tables.LinkColumn(
        "delete_additional_field_by_id",
        args=[A("id")],
        verbose_name=_("Delete"),
        text=_("Delete"),
        attrs={"a": {"class": "btn-flat waves-effect waves-red"}},
    )


class GroupTypesTable(tables.Table):
    """Table to list group types."""

    class Meta:
        attrs = {"class": "responsive-table highlight"}

    name = tables.LinkColumn("edit_group_type_by_id", args=[A("id")])
    description = tables.LinkColumn("edit_group_type_by_id", args=[A("id")])
    delete = tables.LinkColumn(
        "delete_group_type_by_id", args=[A("id")], verbose_name=_("Delete"), text=_("Delete")
    )


class DashboardWidgetTable(tables.Table):
    """Table to list dashboard widgets."""

    class Meta:
        attrs = {"class": "responsive-table highlight"}

    widget_name = tables.Column(accessor="pk")
    title = tables.LinkColumn("edit_dashboard_widget", args=[A("id")])
    active = tables.BooleanColumn(yesno="check,cancel", attrs={"span": {"class": "material-icons"}})
    delete = tables.LinkColumn(
        "delete_dashboard_widget",
        args=[A("id")],
        text=_("Delete"),
        attrs={"a": {"class": "btn-flat waves-effect waves-red red-text"}},
        verbose_name=_("Actions"),
    )

    def render_widget_name(self, value, record):
        return record._meta.verbose_name



from django.conf import settings
from django.core import management

from .celery import app
from .util.notifications import send_notification as _send_notification


@app.task
def send_notification(notification: int, resend: bool = False) -> None:
    """Send a notification object to its recipient.

    :param notification: primary key of the notification object to send
    :param resend: Define whether to also send if the notification was already sent
    """
    _send_notification(notification, resend)


@app.task
def backup_data() -> None:
    """Backup database and media using django-dbbackup."""
    # Assemble command-line options for dbbackup management command
    db_options = []
    if settings.DBBACKUP_COMPRESS_DB:
        db_options.append("-z")
    if settings.DBBACKUP_ENCRYPT_DB:
        db_options.append("-e")
    if settings.DBBACKUP_CLEANUP_DB:
        db_options.append("-c")

    media_options = []
    if settings.DBBACKUP_COMPRESS_MEDIA:
        media_options.append("-z")
    if settings.DBBACKUP_ENCRYPT_MEDIA:
        media_options.append("-e")
    if settings.DBBACKUP_CLEANUP_MEDIA:
        media_options.append("-c")

    # Hand off to dbbackup's management commands
    management.call_command("dbbackup", *db_options)
    management.call_command("mediabackup", *media_options)



"""Utilities and extensions for django_auth_ldap."""

from django.core.exceptions import PermissionDenied

from django_auth_ldap.backend import LDAPBackend as _LDAPBackend


class LDAPBackend(_LDAPBackend):
    default_settings = {"SET_USABLE_PASSWORD": False}

    def authenticate_ldap_user(self, ldap_user, password):
        """Authenticate user against LDAP and set local password if successful.

        Having a local password is needed to make changing passwords easier. In
        order to catch password changes in a universal way and forward them to
        backends (like LDAP, in this case), getting the old password first is
        necessary to authenticate as that user to LDAP.

        We buy the small insecurity of having a hash of the password in the
        Django database in order to not require it to have global admin permissions
        on the LDAP directory.
        """
        user = super().authenticate_ldap_user(ldap_user, password)

        if self.settings.SET_USABLE_PASSWORD:
            if not user:
                # Fail early and do not try other backends
                raise PermissionDenied("LDAP failed to authenticate user")

            # Set a usable password so users can change their LDAP password
            user.set_password(password)
            user.save()

        return user



"""Helpers/overrides for django-allauth."""

from django.conf import settings
from django.http import HttpRequest

from allauth.account.adapter import DefaultAccountAdapter
from allauth.socialaccount.adapter import DefaultSocialAccountAdapter
from oauth2_provider.oauth2_validators import OAuth2Validator

from .core_helpers import get_site_preferences, has_person


class OurSocialAccountAdapter(DefaultSocialAccountAdapter):
    """Customised adapter that recognises other authentication mechanisms."""

    def validate_disconnect(self, account, accounts):
        """Validate whether or not the socialaccount account can be safely disconnected.

        Honours other authentication backends, i.e. ignores unusable passwords if LDAP is used.
        """
        if "django_auth_ldap.backend.LDAPBackend" in settings.AUTHENTICATION_BACKENDS:
            # Ignore upstream validation error as we do not need a usable password
            return None

        # Let upstream decide whether we can disconnect or not
        return super().validate_disconnect(account, accounts)


class OurAccountAdapter(DefaultAccountAdapter):
    """Customised adapter to allow to disable signup."""

    def is_open_for_signup(self, request):
        return get_site_preferences()["auth__signup_enabled"]


class CustomOAuth2Validator(OAuth2Validator):
    def get_additional_claims(self, request):
        django_request = HttpRequest()
        django_request.META = request.headers

        claims = {
            "preferred_username": request.user.username,
        }

        if "profile" in request.scopes:
            if has_person(request.user):
                claims["given_name"] = request.user.person.first_name
                claims["family_name"] = request.user.person.last_name
                claims["profile"] = django_request.build_absolute_uri(
                    request.user.person.get_absolute_url()
                )
                if request.user.person.photo:
                    claims["picture"] = django_request.build_absolute_uri(
                        request.user.person.photo.url
                    )
            else:
                claims["given_name"] = request.user.first_name
                claims["family_name"] = request.user.last_name

        if "email" in request.scopes:
            if has_person(request.user):
                claims["email"] = request.user.person.email
            else:
                claims["email"] = request.user.email

        if "address" in request.scopes and has_person(request.user):
            claims["address"] = {
                "street_address": request.user.person.street
                + " "
                + request.user.person.housenumber,
                "locality": request.user.person.place,
                "postal_code": request.user.person.postal_code,
            }

        return claims



# Materialize colors (without accent, darken and lighten classes)
COLOURS = [
    ("red", "red"),
    ("pink", "pink"),
    ("purple", "purple"),
    ("deep-purple", "deep-purple"),
    ("indigo", "indigo"),
    ("blue", "blue"),
    ("light-blue", "light-blue"),
    ("cyan", "cyan"),
    ("teal", "teal"),
    ("green", "green"),
    ("light-green", "light-green"),
    ("lime", "lime"),
    ("yellow", "yellow"),
    ("amber", "amber"),
    ("orange", "orange"),
    ("deep-orange", "deep-orange"),
    ("brown", "brown"),
    ("grey", "grey"),
    ("blue-grey", "blue-grey"),
    ("black", "black"),
    ("white", "white"),
    ("transparent", "transparent"),
]

# About 1000 icon names from Google's github repository
ICONS = [
    ("3d_rotation", "3d_rotation"),
    ("ac_unit", "ac_unit"),
    ("access_alarm", "access_alarm"),
    ("access_alarms", "access_alarms"),
    ("access_time", "access_time"),
    ("accessibility", "accessibility"),
    ("accessible", "accessible"),
    ("account_balance", "account_balance"),
    ("account_balance_wallet", "account_balance_wallet"),
    ("account_box", "account_box"),
    ("account_circle", "account_circle"),
    ("adb", "adb"),
    ("add", "add"),
    ("add_a_photo", "add_a_photo"),
    ("add_alarm", "add_alarm"),
    ("add_alert", "add_alert"),
    ("add_box", "add_box"),
    ("add_circle", "add_circle"),
    ("add_circle_outline", "add_circle_outline"),
    ("add_location", "add_location"),
    ("add_shopping_cart", "add_shopping_cart"),
    ("add_to_photos", "add_to_photos"),
    ("add_to_queue", "add_to_queue"),
    ("adjust", "adjust"),
    ("airline_seat_flat", "airline_seat_flat"),
    ("airline_seat_flat_angled", "airline_seat_flat_angled"),
    ("airline_seat_individual_suite", "airline_seat_individual_suite"),
    ("airline_seat_legroom_extra", "airline_seat_legroom_extra"),
    ("airline_seat_legroom_normal", "airline_seat_legroom_normal"),
    ("airline_seat_legroom_reduced", "airline_seat_legroom_reduced"),
    ("airline_seat_recline_extra", "airline_seat_recline_extra"),
    ("airline_seat_recline_normal", "airline_seat_recline_normal"),
    ("airplanemode_active", "airplanemode_active"),
    ("airplanemode_inactive", "airplanemode_inactive"),
    ("airplay", "airplay"),
    ("airport_shuttle", "airport_shuttle"),
    ("alarm", "alarm"),
    ("alarm_add", "alarm_add"),
    ("alarm_off", "alarm_off"),
    ("alarm_on", "alarm_on"),
    ("album", "album"),
    ("all_inclusive", "all_inclusive"),
    ("all_out", "all_out"),
    ("android", "android"),
    ("announcement", "announcement"),
    ("apps", "apps"),
    ("archive", "archive"),
    ("arrow_back", "arrow_back"),
    ("arrow_downward", "arrow_downward"),
    ("arrow_drop_down", "arrow_drop_down"),
    ("arrow_drop_down_circle", "arrow_drop_down_circle"),
    ("arrow_drop_up", "arrow_drop_up"),
    ("arrow_forward", "arrow_forward"),
    ("arrow_upward", "arrow_upward"),
    ("art_track", "art_track"),
    ("aspect_ratio", "aspect_ratio"),
    ("assessment", "assessment"),
    ("assignment", "assignment"),
    ("assignment_ind", "assignment_ind"),
    ("assignment_late", "assignment_late"),
    ("assignment_return", "assignment_return"),
    ("assignment_returned", "assignment_returned"),
    ("assignment_turned_in", "assignment_turned_in"),
    ("assistant", "assistant"),
    ("assistant_photo", "assistant_photo"),
    ("attach_file", "attach_file"),
    ("attach_money", "attach_money"),
    ("attachment", "attachment"),
    ("audiotrack", "audiotrack"),
    ("autorenew", "autorenew"),
    ("av_timer", "av_timer"),
    ("backspace", "backspace"),
    ("backup", "backup"),
    ("battery_alert", "battery_alert"),
    ("battery_charging_full", "battery_charging_full"),
    ("battery_full", "battery_full"),
    ("battery_std", "battery_std"),
    ("battery_unknown", "battery_unknown"),
    ("beach_access", "beach_access"),
    ("beenhere", "beenhere"),
    ("block", "block"),
    ("bluetooth", "bluetooth"),
    ("bluetooth_audio", "bluetooth_audio"),
    ("bluetooth_connected", "bluetooth_connected"),
    ("bluetooth_disabled", "bluetooth_disabled"),
    ("bluetooth_searching", "bluetooth_searching"),
    ("blur_circular", "blur_circular"),
    ("blur_linear", "blur_linear"),
    ("blur_off", "blur_off"),
    ("blur_on", "blur_on"),
    ("book", "book"),
    ("bookmark", "bookmark"),
    ("bookmark_border", "bookmark_border"),
    ("border_all", "border_all"),
    ("border_bottom", "border_bottom"),
    ("border_clear", "border_clear"),
    ("border_color", "border_color"),
    ("border_horizontal", "border_horizontal"),
    ("border_inner", "border_inner"),
    ("border_left", "border_left"),
    ("border_outer", "border_outer"),
    ("border_right", "border_right"),
    ("border_style", "border_style"),
    ("border_top", "border_top"),
    ("border_vertical", "border_vertical"),
    ("branding_watermark", "branding_watermark"),
    ("brightness_1", "brightness_1"),
    ("brightness_2", "brightness_2"),
    ("brightness_3", "brightness_3"),
    ("brightness_4", "brightness_4"),
    ("brightness_5", "brightness_5"),
    ("brightness_6", "brightness_6"),
    ("brightness_7", "brightness_7"),
    ("brightness_auto", "brightness_auto"),
    ("brightness_high", "brightness_high"),
    ("brightness_low", "brightness_low"),
    ("brightness_medium", "brightness_medium"),
    ("broken_image", "broken_image"),
    ("brush", "brush"),
    ("bubble_chart", "bubble_chart"),
    ("bug_report", "bug_report"),
    ("build", "build"),
    ("burst_mode", "burst_mode"),
    ("business", "business"),
    ("business_center", "business_center"),
    ("cached", "cached"),
    ("cake", "cake"),
    ("call", "call"),
    ("call_end", "call_end"),
    ("call_made", "call_made"),
    ("call_merge", "call_merge"),
    ("call_missed", "call_missed"),
    ("call_missed_outgoing", "call_missed_outgoing"),
    ("call_received", "call_received"),
    ("call_split", "call_split"),
    ("call_to_action", "call_to_action"),
    ("camera", "camera"),
    ("camera_alt", "camera_alt"),
    ("camera_enhance", "camera_enhance"),
    ("camera_front", "camera_front"),
    ("camera_rear", "camera_rear"),
    ("camera_roll", "camera_roll"),
    ("cancel", "cancel"),
    ("card_giftcard", "card_giftcard"),
    ("card_membership", "card_membership"),
    ("card_travel", "card_travel"),
    ("casino", "casino"),
    ("cast", "cast"),
    ("cast_connected", "cast_connected"),
    ("center_focus_strong", "center_focus_strong"),
    ("center_focus_weak", "center_focus_weak"),
    ("change_history", "change_history"),
    ("chat", "chat"),
    ("chat_bubble", "chat_bubble"),
    ("chat_bubble_outline", "chat_bubble_outline"),
    ("check", "check"),
    ("check_box", "check_box"),
    ("check_box_outline_blank", "check_box_outline_blank"),
    ("check_circle", "check_circle"),
    ("chevron_left", "chevron_left"),
    ("chevron_right", "chevron_right"),
    ("child_care", "child_care"),
    ("child_friendly", "child_friendly"),
    ("chrome_reader_mode", "chrome_reader_mode"),
    ("class", "class"),
    ("clear", "clear"),
    ("clear_all", "clear_all"),
    ("close", "close"),
    ("closed_caption", "closed_caption"),
    ("cloud", "cloud"),
    ("cloud_circle", "cloud_circle"),
    ("cloud_done", "cloud_done"),
    ("cloud_download", "cloud_download"),
    ("cloud_off", "cloud_off"),
    ("cloud_queue", "cloud_queue"),
    ("cloud_upload", "cloud_upload"),
    ("code", "code"),
    ("collections", "collections"),
    ("collections_bookmark", "collections_bookmark"),
    ("color_lens", "color_lens"),
    ("colorize", "colorize"),
    ("comment", "comment"),
    ("compare", "compare"),
    ("compare_arrows", "compare_arrows"),
    ("computer", "computer"),
    ("confirmation_number", "confirmation_number"),
    ("contact_mail", "contact_mail"),
    ("contact_phone", "contact_phone"),
    ("contacts", "contacts"),
    ("content_copy", "content_copy"),
    ("content_cut", "content_cut"),
    ("content_paste", "content_paste"),
    ("control_point", "control_point"),
    ("control_point_duplicate", "control_point_duplicate"),
    ("copyright", "copyright"),
    ("create", "create"),
    ("create_new_folder", "create_new_folder"),
    ("credit_card", "credit_card"),
    ("crop", "crop"),
    ("crop_16_9", "crop_16_9"),
    ("crop_3_2", "crop_3_2"),
    ("crop_5_4", "crop_5_4"),
    ("crop_7_5", "crop_7_5"),
    ("crop_din", "crop_din"),
    ("crop_free", "crop_free"),
    ("crop_landscape", "crop_landscape"),
    ("crop_original", "crop_original"),
    ("crop_portrait", "crop_portrait"),
    ("crop_rotate", "crop_rotate"),
    ("crop_square", "crop_square"),
    ("dashboard", "dashboard"),
    ("data_usage", "data_usage"),
    ("date_range", "date_range"),
    ("dehaze", "dehaze"),
    ("delete", "delete"),
    ("delete_forever", "delete_forever"),
    ("delete_sweep", "delete_sweep"),
    ("description", "description"),
    ("desktop_mac", "desktop_mac"),
    ("desktop_windows", "desktop_windows"),
    ("details", "details"),
    ("developer_board", "developer_board"),
    ("developer_mode", "developer_mode"),
    ("device_hub", "device_hub"),
    ("devices", "devices"),
    ("devices_other", "devices_other"),
    ("dialer_sip", "dialer_sip"),
    ("dialpad", "dialpad"),
    ("directions", "directions"),
    ("directions_bike", "directions_bike"),
    ("directions_boat", "directions_boat"),
    ("directions_bus", "directions_bus"),
    ("directions_car", "directions_car"),
    ("directions_railway", "directions_railway"),
    ("directions_run", "directions_run"),
    ("directions_subway", "directions_subway"),
    ("directions_transit", "directions_transit"),
    ("directions_walk", "directions_walk"),
    ("disc_full", "disc_full"),
    ("dns", "dns"),
    ("do_not_disturb", "do_not_disturb"),
    ("do_not_disturb_alt", "do_not_disturb_alt"),
    ("do_not_disturb_off", "do_not_disturb_off"),
    ("do_not_disturb_on", "do_not_disturb_on"),
    ("dock", "dock"),
    ("domain", "domain"),
    ("done", "done"),
    ("done_all", "done_all"),
    ("donut_large", "donut_large"),
    ("donut_small", "donut_small"),
    ("drafts", "drafts"),
    ("drag_handle", "drag_handle"),
    ("drive_eta", "drive_eta"),
    ("dvr", "dvr"),
    ("edit", "edit"),
    ("edit_location", "edit_location"),
    ("eject", "eject"),
    ("email", "email"),
    ("enhanced_encryption", "enhanced_encryption"),
    ("equalizer", "equalizer"),
    ("error", "error"),
    ("error_outline", "error_outline"),
    ("euro_symbol", "euro_symbol"),
    ("ev_station", "ev_station"),
    ("event", "event"),
    ("event_available", "event_available"),
    ("event_busy", "event_busy"),
    ("event_note", "event_note"),
    ("event_seat", "event_seat"),
    ("exit_to_app", "exit_to_app"),
    ("expand_less", "expand_less"),
    ("expand_more", "expand_more"),
    ("explicit", "explicit"),
    ("explore", "explore"),
    ("exposure", "exposure"),
    ("exposure_neg_1", "exposure_neg_1"),
    ("exposure_neg_2", "exposure_neg_2"),
    ("exposure_plus_1", "exposure_plus_1"),
    ("exposure_plus_2", "exposure_plus_2"),
    ("exposure_zero", "exposure_zero"),
    ("extension", "extension"),
    ("face", "face"),
    ("fast_forward", "fast_forward"),
    ("fast_rewind", "fast_rewind"),
    ("favorite", "favorite"),
    ("favorite_border", "favorite_border"),
    ("featured_play_list", "featured_play_list"),
    ("featured_video", "featured_video"),
    ("feedback", "feedback"),
    ("fiber_dvr", "fiber_dvr"),
    ("fiber_manual_record", "fiber_manual_record"),
    ("fiber_new", "fiber_new"),
    ("fiber_pin", "fiber_pin"),
    ("fiber_smart_record", "fiber_smart_record"),
    ("file_download", "file_download"),
    ("file_upload", "file_upload"),
    ("filter", "filter"),
    ("filter_1", "filter_1"),
    ("filter_2", "filter_2"),
    ("filter_3", "filter_3"),
    ("filter_4", "filter_4"),
    ("filter_5", "filter_5"),
    ("filter_6", "filter_6"),
    ("filter_7", "filter_7"),
    ("filter_8", "filter_8"),
    ("filter_9", "filter_9"),
    ("filter_9_plus", "filter_9_plus"),
    ("filter_b_and_w", "filter_b_and_w"),
    ("filter_center_focus", "filter_center_focus"),
    ("filter_drama", "filter_drama"),
    ("filter_frames", "filter_frames"),
    ("filter_hdr", "filter_hdr"),
    ("filter_list", "filter_list"),
    ("filter_none", "filter_none"),
    ("filter_tilt_shift", "filter_tilt_shift"),
    ("filter_vintage", "filter_vintage"),
    ("find_in_page", "find_in_page"),
    ("find_replace", "find_replace"),
    ("fingerprint", "fingerprint"),
    ("first_page", "first_page"),
    ("fitness_center", "fitness_center"),
    ("flag", "flag"),
    ("flare", "flare"),
    ("flash_auto", "flash_auto"),
    ("flash_off", "flash_off"),
    ("flash_on", "flash_on"),
    ("flight", "flight"),
    ("flight_land", "flight_land"),
    ("flight_takeoff", "flight_takeoff"),
    ("flip", "flip"),
    ("flip_to_back", "flip_to_back"),
    ("flip_to_front", "flip_to_front"),
    ("folder", "folder"),
    ("folder_open", "folder_open"),
    ("folder_shared", "folder_shared"),
    ("folder_special", "folder_special"),
    ("font_download", "font_download"),
    ("format_align_center", "format_align_center"),
    ("format_align_justify", "format_align_justify"),
    ("format_align_left", "format_align_left"),
    ("format_align_right", "format_align_right"),
    ("format_bold", "format_bold"),
    ("format_clear", "format_clear"),
    ("format_color_fill", "format_color_fill"),
    ("format_color_reset", "format_color_reset"),
    ("format_color_text", "format_color_text"),
    ("format_indent_decrease", "format_indent_decrease"),
    ("format_indent_increase", "format_indent_increase"),
    ("format_italic", "format_italic"),
    ("format_line_spacing", "format_line_spacing"),
    ("format_list_bulleted", "format_list_bulleted"),
    ("format_list_numbered", "format_list_numbered"),
    ("format_paint", "format_paint"),
    ("format_quote", "format_quote"),
    ("format_shapes", "format_shapes"),
    ("format_size", "format_size"),
    ("format_strikethrough", "format_strikethrough"),
    ("format_textdirection_l_to_r", "format_textdirection_l_to_r"),
    ("format_textdirection_r_to_l", "format_textdirection_r_to_l"),
    ("format_underlined", "format_underlined"),
    ("forum", "forum"),
    ("forward", "forward"),
    ("forward_10", "forward_10"),
    ("forward_30", "forward_30"),
    ("forward_5", "forward_5"),
    ("free_breakfast", "free_breakfast"),
    ("fullscreen", "fullscreen"),
    ("fullscreen_exit", "fullscreen_exit"),
    ("functions", "functions"),
    ("g_translate", "g_translate"),
    ("gamepad", "gamepad"),
    ("games", "games"),
    ("gavel", "gavel"),
    ("gesture", "gesture"),
    ("get_app", "get_app"),
    ("gif", "gif"),
    ("golf_course", "golf_course"),
    ("gps_fixed", "gps_fixed"),
    ("gps_not_fixed", "gps_not_fixed"),
    ("gps_off", "gps_off"),
    ("grade", "grade"),
    ("gradient", "gradient"),
    ("grain", "grain"),
    ("graphic_eq", "graphic_eq"),
    ("grid_off", "grid_off"),
    ("grid_on", "grid_on"),
    ("group", "group"),
    ("group_add", "group_add"),
    ("group_work", "group_work"),
    ("hd", "hd"),
    ("hdr_off", "hdr_off"),
    ("hdr_on", "hdr_on"),
    ("hdr_strong", "hdr_strong"),
    ("hdr_weak", "hdr_weak"),
    ("headset", "headset"),
    ("headset_mic", "headset_mic"),
    ("healing", "healing"),
    ("hearing", "hearing"),
    ("help", "help"),
    ("help_outline", "help_outline"),
    ("high_quality", "high_quality"),
    ("highlight", "highlight"),
    ("highlight_off", "highlight_off"),
    ("history", "history"),
    ("home", "home"),
    ("hot_tub", "hot_tub"),
    ("hotel", "hotel"),
    ("hourglass_empty", "hourglass_empty"),
    ("hourglass_full", "hourglass_full"),
    ("http", "http"),
    ("https", "https"),
    ("image", "image"),
    ("image_aspect_ratio", "image_aspect_ratio"),
    ("import_contacts", "import_contacts"),
    ("import_export", "import_export"),
    ("important_devices", "important_devices"),
    ("inbox", "inbox"),
    ("indeterminate_check_box", "indeterminate_check_box"),
    ("info", "info"),
    ("info_outline", "info_outline"),
    ("input", "input"),
    ("insert_chart", "insert_chart"),
    ("insert_comment", "insert_comment"),
    ("insert_drive_file", "insert_drive_file"),
    ("insert_emoticon", "insert_emoticon"),
    ("insert_invitation", "insert_invitation"),
    ("insert_link", "insert_link"),
    ("insert_photo", "insert_photo"),
    ("invert_colors", "invert_colors"),
    ("invert_colors_off", "invert_colors_off"),
    ("iso", "iso"),
    ("keyboard", "keyboard"),
    ("keyboard_arrow_down", "keyboard_arrow_down"),
    ("keyboard_arrow_left", "keyboard_arrow_left"),
    ("keyboard_arrow_right", "keyboard_arrow_right"),
    ("keyboard_arrow_up", "keyboard_arrow_up"),
    ("keyboard_backspace", "keyboard_backspace"),
    ("keyboard_capslock", "keyboard_capslock"),
    ("keyboard_hide", "keyboard_hide"),
    ("keyboard_return", "keyboard_return"),
    ("keyboard_tab", "keyboard_tab"),
    ("keyboard_voice", "keyboard_voice"),
    ("kitchen", "kitchen"),
    ("label", "label"),
    ("label_outline", "label_outline"),
    ("landscape", "landscape"),
    ("language", "language"),
    ("laptop", "laptop"),
    ("laptop_chromebook", "laptop_chromebook"),
    ("laptop_mac", "laptop_mac"),
    ("laptop_windows", "laptop_windows"),
    ("last_page", "last_page"),
    ("launch", "launch"),
    ("layers", "layers"),
    ("layers_clear", "layers_clear"),
    ("leak_add", "leak_add"),
    ("leak_remove", "leak_remove"),
    ("lens", "lens"),
    ("library_add", "library_add"),
    ("library_books", "library_books"),
    ("library_music", "library_music"),
    ("lightbulb_outline", "lightbulb_outline"),
    ("line_style", "line_style"),
    ("line_weight", "line_weight"),
    ("linear_scale", "linear_scale"),
    ("link", "link"),
    ("linked_camera", "linked_camera"),
    ("list", "list"),
    ("live_help", "live_help"),
    ("live_tv", "live_tv"),
    ("local_activity", "local_activity"),
    ("local_airport", "local_airport"),
    ("local_atm", "local_atm"),
    ("local_bar", "local_bar"),
    ("local_cafe", "local_cafe"),
    ("local_car_wash", "local_car_wash"),
    ("local_convenience_store", "local_convenience_store"),
    ("local_dining", "local_dining"),
    ("local_drink", "local_drink"),
    ("local_florist", "local_florist"),
    ("local_gas_station", "local_gas_station"),
    ("local_grocery_store", "local_grocery_store"),
    ("local_hospital", "local_hospital"),
    ("local_hotel", "local_hotel"),
    ("local_laundry_service", "local_laundry_service"),
    ("local_library", "local_library"),
    ("local_mall", "local_mall"),
    ("local_movies", "local_movies"),
    ("local_offer", "local_offer"),
    ("local_parking", "local_parking"),
    ("local_pharmacy", "local_pharmacy"),
    ("local_phone", "local_phone"),
    ("local_pizza", "local_pizza"),
    ("local_play", "local_play"),
    ("local_post_office", "local_post_office"),
    ("local_printshop", "local_printshop"),
    ("local_see", "local_see"),
    ("local_shipping", "local_shipping"),
    ("local_taxi", "local_taxi"),
    ("location_city", "location_city"),
    ("location_disabled", "location_disabled"),
    ("location_off", "location_off"),
    ("location_on", "location_on"),
    ("location_searching", "location_searching"),
    ("lock", "lock"),
    ("lock_open", "lock_open"),
    ("lock_outline", "lock_outline"),
    ("looks", "looks"),
    ("looks_3", "looks_3"),
    ("looks_4", "looks_4"),
    ("looks_5", "looks_5"),
    ("looks_6", "looks_6"),
    ("looks_one", "looks_one"),
    ("looks_two", "looks_two"),
    ("loop", "loop"),
    ("loupe", "loupe"),
    ("low_priority", "low_priority"),
    ("loyalty", "loyalty"),
    ("mail", "mail"),
    ("mail_outline", "mail_outline"),
    ("map", "map"),
    ("markunread", "markunread"),
    ("markunread_mailbox", "markunread_mailbox"),
    ("memory", "memory"),
    ("menu", "menu"),
    ("merge_type", "merge_type"),
    ("message", "message"),
    ("mic", "mic"),
    ("mic_none", "mic_none"),
    ("mic_off", "mic_off"),
    ("mms", "mms"),
    ("mode_comment", "mode_comment"),
    ("mode_edit", "mode_edit"),
    ("monetization_on", "monetization_on"),
    ("money_off", "money_off"),
    ("monochrome_photos", "monochrome_photos"),
    ("mood", "mood"),
    ("mood_bad", "mood_bad"),
    ("more", "more"),
    ("more_horiz", "more_horiz"),
    ("more_vert", "more_vert"),
    ("motorcycle", "motorcycle"),
    ("mouse", "mouse"),
    ("move_to_inbox", "move_to_inbox"),
    ("movie", "movie"),
    ("movie_creation", "movie_creation"),
    ("movie_filter", "movie_filter"),
    ("multiline_chart", "multiline_chart"),
    ("music_note", "music_note"),
    ("music_video", "music_video"),
    ("my_location", "my_location"),
    ("nature", "nature"),
    ("nature_people", "nature_people"),
    ("navigate_before", "navigate_before"),
    ("navigate_next", "navigate_next"),
    ("navigation", "navigation"),
    ("near_me", "near_me"),
    ("network_cell", "network_cell"),
    ("network_check", "network_check"),
    ("network_locked", "network_locked"),
    ("network_wifi", "network_wifi"),
    ("new_releases", "new_releases"),
    ("next_week", "next_week"),
    ("nfc", "nfc"),
    ("no_encryption", "no_encryption"),
    ("no_sim", "no_sim"),
    ("not_interested", "not_interested"),
    ("note", "note"),
    ("note_add", "note_add"),
    ("notifications", "notifications"),
    ("notifications_active", "notifications_active"),
    ("notifications_none", "notifications_none"),
    ("notifications_off", "notifications_off"),
    ("notifications_paused", "notifications_paused"),
    ("offline_pin", "offline_pin"),
    ("ondemand_video", "ondemand_video"),
    ("opacity", "opacity"),
    ("open_in_browser", "open_in_browser"),
    ("open_in_new", "open_in_new"),
    ("open_with", "open_with"),
    ("pages", "pages"),
    ("pageview", "pageview"),
    ("palette", "palette"),
    ("pan_tool", "pan_tool"),
    ("panorama", "panorama"),
    ("panorama_fish_eye", "panorama_fish_eye"),
    ("panorama_horizontal", "panorama_horizontal"),
    ("panorama_vertical", "panorama_vertical"),
    ("panorama_wide_angle", "panorama_wide_angle"),
    ("party_mode", "party_mode"),
    ("pause", "pause"),
    ("pause_circle_filled", "pause_circle_filled"),
    ("pause_circle_outline", "pause_circle_outline"),
    ("payment", "payment"),
    ("people", "people"),
    ("people_outline", "people_outline"),
    ("perm_camera_mic", "perm_camera_mic"),
    ("perm_contact_calendar", "perm_contact_calendar"),
    ("perm_data_setting", "perm_data_setting"),
    ("perm_device_information", "perm_device_information"),
    ("perm_identity", "perm_identity"),
    ("perm_media", "perm_media"),
    ("perm_phone_msg", "perm_phone_msg"),
    ("perm_scan_wifi", "perm_scan_wifi"),
    ("person", "person"),
    ("person_add", "person_add"),
    ("person_outline", "person_outline"),
    ("person_pin", "person_pin"),
    ("person_pin_circle", "person_pin_circle"),
    ("personal_video", "personal_video"),
    ("pets", "pets"),
    ("phone", "phone"),
    ("phone_android", "phone_android"),
    ("phone_bluetooth_speaker", "phone_bluetooth_speaker"),
    ("phone_forwarded", "phone_forwarded"),
    ("phone_in_talk", "phone_in_talk"),
    ("phone_iphone", "phone_iphone"),
    ("phone_locked", "phone_locked"),
    ("phone_missed", "phone_missed"),
    ("phone_paused", "phone_paused"),
    ("phonelink", "phonelink"),
    ("phonelink_erase", "phonelink_erase"),
    ("phonelink_lock", "phonelink_lock"),
    ("phonelink_off", "phonelink_off"),
    ("phonelink_ring", "phonelink_ring"),
    ("phonelink_setup", "phonelink_setup"),
    ("photo", "photo"),
    ("photo_album", "photo_album"),
    ("photo_camera", "photo_camera"),
    ("photo_filter", "photo_filter"),
    ("photo_library", "photo_library"),
    ("photo_size_select_actual", "photo_size_select_actual"),
    ("photo_size_select_large", "photo_size_select_large"),
    ("photo_size_select_small", "photo_size_select_small"),
    ("picture_as_pdf", "picture_as_pdf"),
    ("picture_in_picture", "picture_in_picture"),
    ("picture_in_picture_alt", "picture_in_picture_alt"),
    ("pie_chart", "pie_chart"),
    ("pie_chart_outlined", "pie_chart_outlined"),
    ("pin_drop", "pin_drop"),
    ("place", "place"),
    ("play_arrow", "play_arrow"),
    ("play_circle_filled", "play_circle_filled"),
    ("play_circle_outline", "play_circle_outline"),
    ("play_for_work", "play_for_work"),
    ("playlist_add", "playlist_add"),
    ("playlist_add_check", "playlist_add_check"),
    ("playlist_play", "playlist_play"),
    ("plus_one", "plus_one"),
    ("poll", "poll"),
    ("polymer", "polymer"),
    ("pool", "pool"),
    ("portable_wifi_off", "portable_wifi_off"),
    ("portrait", "portrait"),
    ("power", "power"),
    ("power_input", "power_input"),
    ("power_settings_new", "power_settings_new"),
    ("pregnant_woman", "pregnant_woman"),
    ("present_to_all", "present_to_all"),
    ("print", "print"),
    ("priority_high", "priority_high"),
    ("public", "public"),
    ("publish", "publish"),
    ("query_builder", "query_builder"),
    ("question_answer", "question_answer"),
    ("queue", "queue"),
    ("queue_music", "queue_music"),
    ("queue_play_next", "queue_play_next"),
    ("radio", "radio"),
    ("radio_button_checked", "radio_button_checked"),
    ("radio_button_unchecked", "radio_button_unchecked"),
    ("rate_review", "rate_review"),
    ("receipt", "receipt"),
    ("recent_actors", "recent_actors"),
    ("record_voice_over", "record_voice_over"),
    ("redeem", "redeem"),
    ("redo", "redo"),
    ("refresh", "refresh"),
    ("remove", "remove"),
    ("remove_circle", "remove_circle"),
    ("remove_circle_outline", "remove_circle_outline"),
    ("remove_from_queue", "remove_from_queue"),
    ("remove_red_eye", "remove_red_eye"),
    ("remove_shopping_cart", "remove_shopping_cart"),
    ("reorder", "reorder"),
    ("repeat", "repeat"),
    ("repeat_one", "repeat_one"),
    ("replay", "replay"),
    ("replay_10", "replay_10"),
    ("replay_30", "replay_30"),
    ("replay_5", "replay_5"),
    ("reply", "reply"),
    ("reply_all", "reply_all"),
    ("report", "report"),
    ("report_problem", "report_problem"),
    ("restaurant", "restaurant"),
    ("restaurant_menu", "restaurant_menu"),
    ("restore", "restore"),
    ("restore_page", "restore_page"),
    ("ring_volume", "ring_volume"),
    ("room", "room"),
    ("room_service", "room_service"),
    ("rotate_90_degrees_ccw", "rotate_90_degrees_ccw"),
    ("rotate_left", "rotate_left"),
    ("rotate_right", "rotate_right"),
    ("rounded_corner", "rounded_corner"),
    ("router", "router"),
    ("rowing", "rowing"),
    ("rss_feed", "rss_feed"),
    ("rv_hookup", "rv_hookup"),
    ("satellite", "satellite"),
    ("save", "save"),
    ("scanner", "scanner"),
    ("schedule", "schedule"),
    ("school", "school"),
    ("screen_lock_landscape", "screen_lock_landscape"),
    ("screen_lock_portrait", "screen_lock_portrait"),
    ("screen_lock_rotation", "screen_lock_rotation"),
    ("screen_rotation", "screen_rotation"),
    ("screen_share", "screen_share"),
    ("sd_card", "sd_card"),
    ("sd_storage", "sd_storage"),
    ("search", "search"),
    ("security", "security"),
    ("select_all", "select_all"),
    ("send", "send"),
    ("sentiment_dissatisfied", "sentiment_dissatisfied"),
    ("sentiment_neutral", "sentiment_neutral"),
    ("sentiment_satisfied", "sentiment_satisfied"),
    ("sentiment_very_dissatisfied", "sentiment_very_dissatisfied"),
    ("sentiment_very_satisfied", "sentiment_very_satisfied"),
    ("settings", "settings"),
    ("settings_applications", "settings_applications"),
    ("settings_backup_restore", "settings_backup_restore"),
    ("settings_bluetooth", "settings_bluetooth"),
    ("settings_brightness", "settings_brightness"),
    ("settings_cell", "settings_cell"),
    ("settings_ethernet", "settings_ethernet"),
    ("settings_input_antenna", "settings_input_antenna"),
    ("settings_input_component", "settings_input_component"),
    ("settings_input_composite", "settings_input_composite"),
    ("settings_input_hdmi", "settings_input_hdmi"),
    ("settings_input_svideo", "settings_input_svideo"),
    ("settings_overscan", "settings_overscan"),
    ("settings_phone", "settings_phone"),
    ("settings_power", "settings_power"),
    ("settings_remote", "settings_remote"),
    ("settings_system_daydream", "settings_system_daydream"),
    ("settings_voice", "settings_voice"),
    ("share", "share"),
    ("shop", "shop"),
    ("shop_two", "shop_two"),
    ("shopping_basket", "shopping_basket"),
    ("shopping_cart", "shopping_cart"),
    ("short_text", "short_text"),
    ("show_chart", "show_chart"),
    ("shuffle", "shuffle"),
    ("signal_cellular_4_bar", "signal_cellular_4_bar"),
    ("signal_cellular_connected_no_internet_4_bar", "signal_cellular_connected_no_internet_4_bar"),
    ("signal_cellular_no_sim", "signal_cellular_no_sim"),
    ("signal_cellular_null", "signal_cellular_null"),
    ("signal_cellular_off", "signal_cellular_off"),
    ("signal_wifi_4_bar", "signal_wifi_4_bar"),
    ("signal_wifi_4_bar_lock", "signal_wifi_4_bar_lock"),
    ("signal_wifi_off", "signal_wifi_off"),
    ("sim_card", "sim_card"),
    ("sim_card_alert", "sim_card_alert"),
    ("skip_next", "skip_next"),
    ("skip_previous", "skip_previous"),
    ("slideshow", "slideshow"),
    ("slow_motion_video", "slow_motion_video"),
    ("smartphone", "smartphone"),
    ("smoke_free", "smoke_free"),
    ("smoking_rooms", "smoking_rooms"),
    ("sms", "sms"),
    ("sms_failed", "sms_failed"),
    ("snooze", "snooze"),
    ("sort", "sort"),
    ("sort_by_alpha", "sort_by_alpha"),
    ("spa", "spa"),
    ("space_bar", "space_bar"),
    ("speaker", "speaker"),
    ("speaker_group", "speaker_group"),
    ("speaker_notes", "speaker_notes"),
    ("speaker_notes_off", "speaker_notes_off"),
    ("speaker_phone", "speaker_phone"),
    ("spellcheck", "spellcheck"),
    ("star", "star"),
    ("star_border", "star_border"),
    ("star_half", "star_half"),
    ("stars", "stars"),
    ("stay_current_landscape", "stay_current_landscape"),
    ("stay_current_portrait", "stay_current_portrait"),
    ("stay_primary_landscape", "stay_primary_landscape"),
    ("stay_primary_portrait", "stay_primary_portrait"),
    ("stop", "stop"),
    ("stop_screen_share", "stop_screen_share"),
    ("storage", "storage"),
    ("store", "store"),
    ("store_mall_directory", "store_mall_directory"),
    ("straighten", "straighten"),
    ("streetview", "streetview"),
    ("strikethrough_s", "strikethrough_s"),
    ("style", "style"),
    ("subdirectory_arrow_left", "subdirectory_arrow_left"),
    ("subdirectory_arrow_right", "subdirectory_arrow_right"),
    ("subject", "subject"),
    ("subscriptions", "subscriptions"),
    ("subtitles", "subtitles"),
    ("subway", "subway"),
    ("supervisor_account", "supervisor_account"),
    ("surround_sound", "surround_sound"),
    ("swap_calls", "swap_calls"),
    ("swap_horiz", "swap_horiz"),
    ("swap_vert", "swap_vert"),
    ("swap_vertical_circle", "swap_vertical_circle"),
    ("switch_camera", "switch_camera"),
    ("switch_video", "switch_video"),
    ("sync", "sync"),
    ("sync_disabled", "sync_disabled"),
    ("sync_problem", "sync_problem"),
    ("system_update", "system_update"),
    ("system_update_alt", "system_update_alt"),
    ("tab", "tab"),
    ("tab_unselected", "tab_unselected"),
    ("tablet", "tablet"),
    ("tablet_android", "tablet_android"),
    ("tablet_mac", "tablet_mac"),
    ("tag_faces", "tag_faces"),
    ("tap_and_play", "tap_and_play"),
    ("terrain", "terrain"),
    ("text_fields", "text_fields"),
    ("text_format", "text_format"),
    ("textsms", "textsms"),
    ("texture", "texture"),
    ("theaters", "theaters"),
    ("thumb_down", "thumb_down"),
    ("thumb_up", "thumb_up"),
    ("thumbs_up_down", "thumbs_up_down"),
    ("time_to_leave", "time_to_leave"),
    ("timelapse", "timelapse"),
    ("timeline", "timeline"),
    ("timer", "timer"),
    ("timer_10", "timer_10"),
    ("timer_3", "timer_3"),
    ("timer_off", "timer_off"),
    ("title", "title"),
    ("toc", "toc"),
    ("today", "today"),
    ("toll", "toll"),
    ("tonality", "tonality"),
    ("touch_app", "touch_app"),
    ("toys", "toys"),
    ("track_changes", "track_changes"),
    ("traffic", "traffic"),
    ("train", "train"),
    ("tram", "tram"),
    ("transfer_within_a_station", "transfer_within_a_station"),
    ("transform", "transform"),
    ("translate", "translate"),
    ("trending_down", "trending_down"),
    ("trending_flat", "trending_flat"),
    ("trending_up", "trending_up"),
    ("tune", "tune"),
    ("turned_in", "turned_in"),
    ("turned_in_not", "turned_in_not"),
    ("tv", "tv"),
    ("unarchive", "unarchive"),
    ("undo", "undo"),
    ("unfold_less", "unfold_less"),
    ("unfold_more", "unfold_more"),
    ("update", "update"),
    ("usb", "usb"),
    ("verified_user", "verified_user"),
    ("vertical_align_bottom", "vertical_align_bottom"),
    ("vertical_align_center", "vertical_align_center"),
    ("vertical_align_top", "vertical_align_top"),
    ("vibration", "vibration"),
    ("video_call", "video_call"),
    ("video_label", "video_label"),
    ("video_library", "video_library"),
    ("videocam", "videocam"),
    ("videocam_off", "videocam_off"),
    ("videogame_asset", "videogame_asset"),
    ("view_agenda", "view_agenda"),
    ("view_array", "view_array"),
    ("view_carousel", "view_carousel"),
    ("view_column", "view_column"),
    ("view_comfy", "view_comfy"),
    ("view_compact", "view_compact"),
    ("view_day", "view_day"),
    ("view_headline", "view_headline"),
    ("view_list", "view_list"),
    ("view_module", "view_module"),
    ("view_quilt", "view_quilt"),
    ("view_stream", "view_stream"),
    ("view_week", "view_week"),
    ("vignette", "vignette"),
    ("visibility", "visibility"),
    ("visibility_off", "visibility_off"),
    ("voice_chat", "voice_chat"),
    ("voicemail", "voicemail"),
    ("volume_down", "volume_down"),
    ("volume_mute", "volume_mute"),
    ("volume_off", "volume_off"),
    ("volume_up", "volume_up"),
    ("vpn_key", "vpn_key"),
    ("vpn_lock", "vpn_lock"),
    ("wallpaper", "wallpaper"),
    ("warning", "warning"),
    ("watch", "watch"),
    ("watch_later", "watch_later"),
    ("wb_auto", "wb_auto"),
    ("wb_cloudy", "wb_cloudy"),
    ("wb_incandescent", "wb_incandescent"),
    ("wb_iridescent", "wb_iridescent"),
    ("wb_sunny", "wb_sunny"),
    ("wc", "wc"),
    ("web", "web"),
    ("web_asset", "web_asset"),
    ("weekend", "weekend"),
    ("whatshot", "whatshot"),
    ("widgets", "widgets"),
    ("wifi", "wifi"),
    ("wifi_lock", "wifi_lock"),
    ("wifi_tethering", "wifi_tethering"),
    ("work", "work"),
    ("wrap_text", "wrap_text"),
    ("youtube_searched_for", "youtube_searched_for"),
    ("zoom_in", "zoom_in"),
    ("zoom_out", "zoom_out"),
    ("zoom_out_map", "zoom_out_map"),
]



from django.conf import settings
from django.utils.translation import gettext_lazy as _

from .util.core_helpers import unread_notifications_badge

MENUS = {
    "NAV_MENU_CORE": [
        {
            "name": _("Login"),
            "url": settings.LOGIN_URL,
            "icon": "lock_open",
            "validators": ["menu_generator.validators.is_anonymous"],
        },
        {
            "name": _("Sign up"),
            "url": "account_signup",
            "icon": "how_to_reg",
            "validators": [
                "menu_generator.validators.is_anonymous",
                ("aleksis.core.util.predicates.permission_validator", "core.can_register"),
            ],
        },
        {
            "name": _("Dashboard"),
            "url": "index",
            "icon": "home",
            "validators": [
                ("aleksis.core.util.predicates.permission_validator", "core.view_dashboard_rule")
            ],
        },
        {
            "name": _("Notifications"),
            "url": "notifications",
            "icon": "notifications",
            "badge": unread_notifications_badge,
            "validators": [
                ("aleksis.core.util.predicates.permission_validator", "core.view_notifications",),
            ],
        },
        {
            "name": _("Account"),
            "url": "#",
            "icon": "person",
            "root": True,
            "validators": ["menu_generator.validators.is_authenticated"],
            "submenu": [
                {
                    "name": _("Stop impersonation"),
                    "url": "impersonate-stop",
                    "icon": "stop",
                    "validators": [
                        "menu_generator.validators.is_authenticated",
                        "aleksis.core.util.core_helpers.is_impersonate",
                    ],
                },
                {
                    "name": _("Logout"),
                    "url": "logout",
                    "icon": "exit_to_app",
                    "validators": ["menu_generator.validators.is_authenticated"],
                },
                {
                    "name": _("2FA"),
                    "url": "two_factor:profile",
                    "icon": "phonelink_lock",
                    "validators": ["menu_generator.validators.is_authenticated",],
                },
                {
                    "name": _("Change password"),
                    "url": "account_change_password",
                    "icon": "lock",
                    "validators": [
                        "menu_generator.validators.is_authenticated",
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.can_change_password",
                        ),
                    ],
                },
                {
                    "name": _("Me"),
                    "url": "person",
                    "icon": "insert_emoticon",
                    "validators": [
                        "menu_generator.validators.is_authenticated",
                        "aleksis.core.util.core_helpers.has_person",
                    ],
                },
                {
                    "name": _("Preferences"),
                    "url": "preferences_person",
                    "icon": "settings",
                    "validators": [
                        "menu_generator.validators.is_authenticated",
                        "aleksis.core.util.core_helpers.has_person",
                    ],
                },
                {
                    "name": _("Third-party accounts"),
                    "url": "socialaccount_connections",
                    "icon": "public",
                    "validators": [
                        "menu_generator.validators.is_authenticated",
                        "aleksis.core.util.core_helpers.has_person",
                    ],
                },
                {
                    "name": _("Authorized applications"),
                    "url": "oauth2_provider:authorized-token-list",
                    "icon": "touch_app",
                    "validators": [
                        "menu_generator.validators.is_authenticated",
                        "aleksis.core.util.core_helpers.has_person",
                    ],
                },
            ],
        },
        {
            "name": _("Admin"),
            "url": "#",
            "icon": "security",
            "validators": [
                ("aleksis.core.util.predicates.permission_validator", "core.view_admin_menu"),
            ],
            "submenu": [
                {
                    "name": _("Announcements"),
                    "url": "announcements",
                    "icon": "announcement",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.view_announcements_rule",
                        ),
                    ],
                },
                {
                    "name": _("School terms"),
                    "url": "school_terms",
                    "icon": "date_range",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.view_schoolterm_rule",
                        ),
                    ],
                },
                {
                    "name": _("Dashboard widgets"),
                    "url": "dashboard_widgets",
                    "icon": "dashboard",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.view_dashboardwidget_rule",
                        ),
                    ],
                },
                {
                    "name": _("Data management"),
                    "url": "data_management",
                    "icon": "view_list",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.manage_data_rule",
                        ),
                    ],
                },
                {
                    "name": _("System status"),
                    "url": "system_status",
                    "icon": "power_settings_new",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.view_system_status_rule",
                        ),
                    ],
                },
                {
                    "name": _("Impersonation"),
                    "url": "impersonate-list",
                    "icon": "people",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.impersonate_rule",
                        ),
                    ],
                },
                {
                    "name": _("Configuration"),
                    "url": "preferences_site",
                    "icon": "settings",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.change_site_preferences_rule",
                        ),
                    ],
                },
                {
                    "name": _("Data checks"),
                    "url": "check_data",
                    "icon": "done_all",
                    "validators": ["menu_generator.validators.is_superuser"],
                },
                {
                    "name": _("Backend Admin"),
                    "url": "admin:index",
                    "icon": "settings",
                    "validators": ["menu_generator.validators.is_superuser",],
                },
                {
                    "name": _("OAuth2 Applications"),
                    "url": "oauth_list",
                    "icon": "touch_app",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.list_oauth_applications_rule",
                        ),
                    ],
                },
            ],
        },
        {
            "name": _("People"),
            "url": "#",
            "icon": "people",
            "root": True,
            "validators": [
                ("aleksis.core.util.predicates.permission_validator", "core.view_people_menu_rule")
            ],
            "submenu": [
                {
                    "name": _("Persons"),
                    "url": "persons",
                    "icon": "person",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.view_persons_rule",
                        )
                    ],
                },
                {
                    "name": _("Groups"),
                    "url": "groups",
                    "icon": "group",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.view_groups_rule",
                        )
                    ],
                },
                {
                    "name": _("Group types"),
                    "url": "group_types",
                    "icon": "category",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.view_grouptypes_rule",
                        )
                    ],
                },
                {
                    "name": _("Persons and accounts"),
                    "url": "persons_accounts",
                    "icon": "person_add",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.link_persons_accounts_rule",
                        )
                    ],
                },
                {
                    "name": _("Groups and child groups"),
                    "url": "groups_child_groups",
                    "icon": "group_add",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.assign_child_groups_to_groups_rule",
                        )
                    ],
                },
                {
                    "name": _("Additional fields"),
                    "url": "additional_fields",
                    "icon": "style",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.view_additionalfields_rule",
                        )
                    ],
                },
            ],
        },
    ],
    "DATA_MANAGEMENT_MENU": [
        {
            "name": _("Assign child groups to groups"),
            "url": "groups_child_groups",
            "validators": [
                (
                    "aleksis.core.util.predicates.permission_validator",
                    "core.assign_child_groups_to_groups_rule",
                )
            ],
        },
    ],
}



"""Utility code for notification system."""

from typing import Sequence, Union

from django.apps import apps
from django.conf import settings
from django.template.loader import get_template
from django.utils.functional import lazy
from django.utils.translation import gettext_lazy as _

from templated_email import send_templated_mail

from .core_helpers import lazy_preference

try:
    from twilio.rest import Client as TwilioClient
except ImportError:
    TwilioClient = None


def send_templated_sms(
    template_name: str, from_number: str, recipient_list: Sequence[str], context: dict
) -> None:
    """Render a plan-text template and send via SMS to all recipients."""
    template = get_template(template_name)
    text = template.render(context)

    client = TwilioClient(settings.TWILIO_SID, settings.TWILIO_TOKEN)
    for recipient in recipient_list:
        client.messages.create(body=text, to=recipient, from_=from_number)


def _send_notification_email(notification: "Notification", template: str = "notification") -> None:
    context = {
        "notification": notification,
        "notification_user": notification.recipient.addressing_name,
    }
    send_templated_mail(
        template_name=template,
        from_email=lazy_preference("mail", "address"),
        recipient_list=[notification.recipient.email],
        context=context,
    )


def _send_notification_sms(
    notification: "Notification", template: str = "sms/notification.txt"
) -> None:
    context = {
        "notification": notification,
        "notification_user": notification.recipient.addressing_name,
    }
    send_templated_sms(
        template_name=template,
        from_number=settings.TWILIO_CALLER_ID,
        recipient_list=[notification.recipient.mobile_number.as_e164],
        context=context,
    )


# Mapping of channel id to name and two functions:
# - Check for availability
# - Send notification through it
_CHANNELS_MAP = {
    "email": (_("E-Mail"), lambda: lazy_preference("mail", "address"), _send_notification_email),
    "sms": (_("SMS"), lambda: getattr(settings, "TWILIO_SID", None), _send_notification_sms),
}


def send_notification(notification: Union[int, "Notification"], resend: bool = False) -> None:
    """Send a notification through enabled channels.

    If resend is passed as True, the notification is sent even if it was
    previously marked as sent.
    """
    if isinstance(notification, int):
        Notification = apps.get_model("core", "Notification")
        notification = Notification.objects.get(pk=notification)

    channels = [notification.recipient.preferences["notification__channels"]]

    if resend or not notification.sent:
        for channel in channels:
            name, check, send = _CHANNELS_MAP[channel]
            if check():
                send(notification)


def get_notification_choices() -> list:
    """Return all available channels for notifications.

    This gathers the channels that are technically available as per the
    system configuration. Which ones are available to users is defined
    by the administrator (by selecting a subset of these choices).
    """
    choices = []
    for channel, (name, check, send) in _CHANNELS_MAP.items():
        if check():
            choices.append((channel, name))
    return choices


get_notification_choices_lazy = lazy(get_notification_choices, tuple)



from django.conf import settings
from django.urls import reverse

import pytest

pytestmark = pytest.mark.django_db


def test_index_not_logged_in(client):
    response = client.get("/")

    assert response.status_code == 302
    assert response["Location"].startswith(reverse(settings.LOGIN_URL))


def test_login_without_person(client, django_user_model):
    username = "foo"
    password = "bar"

    django_user_model.objects.create_user(username=username, password=password)
    client.login(username=username, password=password)

    response = client.get("/", follow=True)

    assert response.status_code == 200
    assert "Your user account is not linked to a person." in response.content.decode("utf-8")


def test_logout(client, django_user_model):
    username = "foo"
    password = "bar"

    django_user_model.objects.create_user(username=username, password=password)
    client.login(username=username, password=password)

    response = client.get("/", follow=True)
    assert response.status_code == 200

    response = client.get(reverse("logout"), follow=True)

    assert response.status_code == 200
    assert "Please login to see this page." in response.content.decode("utf-8")



import os
from glob import glob
from socket import getfqdn

from django.utils.translation import gettext_lazy as _

from dynaconf import LazySettings

from .util.core_helpers import get_app_packages, merge_app_settings, monkey_patch

monkey_patch()

IN_PYTEST = "PYTEST_CURRENT_TEST" in os.environ or "TOX_ENV_DIR" in os.environ

ENVVAR_PREFIX_FOR_DYNACONF = "ALEKSIS"
DIRS_FOR_DYNACONF = ["/etc/aleksis"]

SETTINGS_FILE_FOR_DYNACONF = []
for directory in DIRS_FOR_DYNACONF:
    SETTINGS_FILE_FOR_DYNACONF += glob(os.path.join(directory, "*.ini"))
    SETTINGS_FILE_FOR_DYNACONF += glob(os.path.join(directory, "*.yaml"))
    SETTINGS_FILE_FOR_DYNACONF += glob(os.path.join(directory, "*.toml"))

_settings = LazySettings(
    ENVVAR_PREFIX_FOR_DYNACONF=ENVVAR_PREFIX_FOR_DYNACONF,
    SETTINGS_FILE_FOR_DYNACONF=SETTINGS_FILE_FOR_DYNACONF,
)

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

SILENCED_SYSTEM_CHECKS = []

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = _settings.get("secret_key", "DoNotUseInProduction")

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = _settings.get("maintenance.debug", False)
INTERNAL_IPS = _settings.get("maintenance.internal_ips", [])
DEBUG_TOOLBAR_CONFIG = {
    "RENDER_PANELS": True,
    "SHOW_COLLAPSED": True,
    "JQUERY_URL": "",
    "SHOW_TOOLBAR_CALLBACK": "aleksis.core.util.core_helpers.dt_show_toolbar",
    "DISABLE_PANELS": {},
}

DEBUG_TOOLBAR_PANELS = [
    "debug_toolbar.panels.versions.VersionsPanel",
    "debug_toolbar.panels.timer.TimerPanel",
    "debug_toolbar.panels.settings.SettingsPanel",
    "debug_toolbar.panels.headers.HeadersPanel",
    "debug_toolbar.panels.request.RequestPanel",
    "debug_toolbar.panels.sql.SQLPanel",
    "debug_toolbar.panels.cache.CachePanel",
    "debug_toolbar.panels.staticfiles.StaticFilesPanel",
    "debug_toolbar.panels.templates.TemplatesPanel",
    "debug_toolbar.panels.signals.SignalsPanel",
    "debug_toolbar.panels.logging.LoggingPanel",
    "debug_toolbar.panels.profiling.ProfilingPanel",
    "django_uwsgi.panels.UwsgiPanel",
]

UWSGI = {
    "module": "aleksis.core.wsgi",
}
UWSGI_SERVE_STATIC = True
UWSGI_SERVE_MEDIA = False

ALLOWED_HOSTS = _settings.get("http.allowed_hosts", [getfqdn(), "localhost", "127.0.0.1", "[::1]"])
BASE_URL = _settings.get(
    "http.base_url", "http://localhost:8000" if DEBUG else f"https://{ALLOWED_HOSTS[0]}"
)

# Application definition
INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.sites",
    "django.contrib.staticfiles",
    "django.contrib.humanize",
    "django_uwsgi",
    "django_extensions",
    "guardian",
    "rules.apps.AutodiscoverRulesConfig",
    "haystack",
    "polymorphic",
    "dbbackup",
    "django_celery_beat",
    "django_celery_results",
    "celery_progress",
    "health_check.contrib.celery",
    "djcelery_email",
    "celery_haystack",
    "sass_processor",
    "django_any_js",
    "django_yarnpkg",
    "django_tables2",
    "maintenance_mode",
    "menu_generator",
    "reversion",
    "phonenumber_field",
    "debug_toolbar",
    "django_prometheus",
    "django_select2",
    "hattori",
    "templated_email",
    "html2text",
    "django_otp.plugins.otp_totp",
    "django_otp.plugins.otp_static",
    "django_otp",
    "otp_yubikey",
    "aleksis.core",
    "allauth",
    "allauth.account",
    "allauth.socialaccount",
    "health_check",
    "health_check.db",
    "health_check.cache",
    "health_check.storage",
    "health_check.contrib.psutil",
    "health_check.contrib.migrations",
    "dynamic_preferences",
    "dynamic_preferences.users.apps.UserPreferencesConfig",
    "impersonate",
    "two_factor",
    "material",
    "ckeditor",
    "ckeditor_uploader",
    "django_js_reverse",
    "colorfield",
    "django_bleach",
    "favicon",
    "django_filters",
    "oauth2_provider",
    "rest_framework",
]

merge_app_settings("INSTALLED_APPS", INSTALLED_APPS, True)
INSTALLED_APPS += get_app_packages()

STATICFILES_FINDERS = [
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
    "django_yarnpkg.finders.NodeModulesFinder",
    "sass_processor.finders.CssFinder",
]

MIDDLEWARE = [
    #    'django.middleware.cache.UpdateCacheMiddleware',
    "django_prometheus.middleware.PrometheusBeforeMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "debug_toolbar.middleware.DebugToolbarMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.http.ConditionalGetMiddleware",
    "django.contrib.sites.middleware.CurrentSiteMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django_otp.middleware.OTPMiddleware",
    "impersonate.middleware.ImpersonateMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "maintenance_mode.middleware.MaintenanceModeMiddleware",
    "aleksis.core.util.middlewares.EnsurePersonMiddleware",
    "django_prometheus.middleware.PrometheusAfterMiddleware",
    #    'django.middleware.cache.FetchFromCacheMiddleware'
]

ROOT_URLCONF = "aleksis.core.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "maintenance_mode.context_processors.maintenance_mode",
                "dynamic_preferences.processors.global_preferences",
                "aleksis.core.util.core_helpers.custom_information_processor",
            ],
        },
    },
]

WSGI_APPLICATION = "aleksis.core.wsgi.application"

# Database
# https://docs.djangoproject.com/en/2.1/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django_prometheus.db.backends.postgresql",
        "NAME": _settings.get("database.name", "aleksis"),
        "USER": _settings.get("database.username", "aleksis"),
        "PASSWORD": _settings.get("database.password", None),
        "HOST": _settings.get("database.host", "127.0.0.1"),
        "PORT": _settings.get("database.port", "5432"),
        "CONN_MAX_AGE": _settings.get("database.conn_max_age", None),
    }
}

merge_app_settings("DATABASES", DATABASES, False)

REDIS_HOST = _settings.get("redis.host", "localhost")
REDIS_PORT = _settings.get("redis.port", 6379)
REDIS_DB = _settings.get("redis.database", 0)
REDIS_PASSWORD = _settings.get("redis.password", None)
REDIS_USER = _settings.get("redis.user", None if REDIS_PASSWORD is None else "default")

REDIS_URL = (
    f"redis://{REDIS_USER+':'+REDIS_PASSWORD+'@' if REDIS_USER else ''}"
    f"{REDIS_HOST}:{REDIS_PORT}/{REDIS_DB}"
)

if _settings.get("caching.redis.enabled", not IN_PYTEST):
    CACHES = {
        "default": {
            "BACKEND": "django_redis.cache.RedisCache",
            "LOCATION": _settings.get("caching.redis.address", REDIS_URL),
            "OPTIONS": {"CLIENT_CLASS": "django_redis.client.DefaultClient",},
        }
    }
    if REDIS_PASSWORD:
        CACHES["default"]["OPTIONS"]["PASSWORD"] = REDIS_PASSWORD
else:
    CACHES = {
        "default": {
            # Use uWSGI if available (will auot-fallback to LocMemCache)
            "BACKEND": "django_uwsgi.cache.UwsgiCache"
        }
    }

INSTALLED_APPS.append("cachalot")
DEBUG_TOOLBAR_PANELS.append("cachalot.panels.CachalotPanel")
CACHALOT_TIMEOUT = _settings.get("caching.cachalot.timeout", None)
CACHALOT_DATABASES = set(["default"])
SILENCED_SYSTEM_CHECKS += ["cachalot.W001"]

SESSION_ENGINE = "django.contrib.sessions.backends.cache"
SESSION_CACHE_ALIAS = "default"

# Password validation
# https://docs.djangoproject.com/en/2.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {"NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",},
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",},
]

AUTH_INITIAL_SUPERUSER = {
    "username": _settings.get("auth.superuser.username", "admin"),
    "password": _settings.get("auth.superuser.password", "admin"),
    "email": _settings.get("auth.superuser.email", "root@example.com"),
}

# Authentication backends are dynamically populated
AUTHENTICATION_BACKENDS = []

# Configuration for django-allauth.

# Use custom adapter to override some behaviour, i.e. honour the LDAP backend
SOCIALACCOUNT_ADAPTER = "aleksis.core.util.auth_helpers.OurSocialAccountAdapter"

# Get django-allauth providers from config
_SOCIALACCOUNT_PROVIDERS = _settings.get("auth.providers", None)
if _SOCIALACCOUNT_PROVIDERS:
    SOCIALACCOUNT_PROVIDERS = _SOCIALACCOUNT_PROVIDERS.to_dict()

    # Add configured social auth providers to INSTALLED_APPS
    for provider, config in SOCIALACCOUNT_PROVIDERS.items():
        INSTALLED_APPS.append(f"allauth.socialaccount.providers.{provider}")
        SOCIALACCOUNT_PROVIDERS[provider] = {k.upper(): v for k, v in config.items()}


# Configure custom forms

ACCOUNT_FORMS = {
    "signup": "aleksis.core.forms.AccountRegisterForm",
}

# Use custom adapter
ACCOUNT_ADAPTER = "aleksis.core.util.auth_helpers.OurAccountAdapter"

# Require password confirmation
SIGNUP_PASSWORD_ENTER_TWICE = True

# Allow login by either username or email
ACCOUNT_AUTHENTICATION_METHOD = _settings.get("auth.registration.method", "username_email")

# Require email address to sign up
ACCOUNT_EMAIL_REQUIRED = _settings.get("auth.registration.email_required", True)
SOCIALACCOUNT_EMAIL_REQUIRED = False

# Require email verification after sigm up
ACCOUNT_EMAIL_VERIFICATION = _settings.get("auth.registration.email_verification", "mandatory")
SOCIALACCOUNT_EMAIL_VERIFICATION = False

# Email subject prefix for verification mails
ACCOUNT_EMAIL_SUBJECT_PREFIX = _settings.get("auth.registration.subject", "[AlekSIS] ")

# Max attempts before login timeout
ACCOUNT_LOGIN_ATTEMPTS_LIMIT = _settings.get("auth.login.login_limit", 5)

# Login timeout after max attempts in seconds
ACCOUNT_LOGIN_ATTEMPTS_TIMEOUT = _settings.get("auth.login.login_timeout", 300)

# Email confirmation field in form
ACCOUNT_SIGNUP_EMAIL_ENTER_TWICE = True

# Enforce uniqueness of email addresses
ACCOUNT_UNIQUE_EMAIL = _settings.get("auth.login.registration.unique_email", True)

# Configuration for OAuth2 provider

OAUTH2_PROVIDER = {
    "SCOPES": {
        "read": "Read anything the resource owner can read",
        "write": "Write anything the resource owner can write",
    }
}
merge_app_settings("OAUTH2_SCOPES", OAUTH2_PROVIDER["SCOPES"], True)

if _settings.get("oauth2.oidc.enabled", False):
    with open(_settings.get("oauth2.oidc.rsa_key", "/etc/aleksis/oidc.pem"), "r") as f:
        oid_rsa_key = f.read()

    OAUTH2_PROVIDER.update(
        {
            "OAUTH2_VALIDATOR_CLASS": "aleksis.core.util.auth_helpers.CustomOAuth2Validator",
            "OIDC_ENABLED": True,
            "OIDC_RSA_PRIVATE_KEY": oid_rsa_key,
            #        "OIDC_ISS_ENDPOINT": _settings.get("oauth2.oidc.issuer_name", "example.com"),
        }
    )
    OAUTH2_PROVIDER["SCOPES"].update(
        {
            "openid": _("OpenID Connect scope"),
            "profile": _("Given name, family name, link to profile and picture if existing."),
            "address": _("Full home postal address"),
            "email": _("Email address"),
            "phone": _("Home and mobile phone"),
        }
    )

# Configuration for REST framework
REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "oauth2_provider.contrib.rest_framework.OAuth2Authentication",
    ]
}

# LDAP config
if _settings.get("ldap.uri", None):
    # LDAP dependencies are not necessarily installed, so import them here
    import ldap  # noqa
    from django_auth_ldap.config import (
        LDAPSearch,
        LDAPSearchUnion,
        NestedGroupOfNamesType,
        NestedGroupOfUniqueNamesType,
        PosixGroupType,
    )

    # Enable Django's integration to LDAP
    AUTHENTICATION_BACKENDS.append("aleksis.core.util.ldap.LDAPBackend")

    AUTH_LDAP_SERVER_URI = _settings.get("ldap.uri")

    # Optional: non-anonymous bind
    if _settings.get("ldap.bind.dn", None):
        AUTH_LDAP_BIND_DN = _settings.get("ldap.bind.dn")
        AUTH_LDAP_BIND_PASSWORD = _settings.get("ldap.bind.password")

    # Keep local password for users to be required to proveide their old password on change
    AUTH_LDAP_SET_USABLE_PASSWORD = _settings.get("ldap.handle_passwords", True)

    # Keep bound as the authenticating user
    # Ensures proper read permissions, and ability to change password without admin
    AUTH_LDAP_BIND_AS_AUTHENTICATING_USER = True

    # The TOML config might contain either one table or an array of tables
    _AUTH_LDAP_USER_SETTINGS = _settings.get("ldap.users.search")
    if not isinstance(_AUTH_LDAP_USER_SETTINGS, list):
        _AUTH_LDAP_USER_SETTINGS = [_AUTH_LDAP_USER_SETTINGS]

    # Search attributes to find users by username
    AUTH_LDAP_USER_SEARCH = LDAPSearchUnion(
        *[
            LDAPSearch(entry["base"], ldap.SCOPE_SUBTREE, entry.get("filter", "(uid=%(user)s)"),)
            for entry in _AUTH_LDAP_USER_SETTINGS
        ]
    )

    # Mapping of LDAP attributes to Django model fields
    AUTH_LDAP_USER_ATTR_MAP = {
        "first_name": _settings.get("ldap.users.map.first_name", "givenName"),
        "last_name": _settings.get("ldap.users.map.last_name", "sn"),
        "email": _settings.get("ldap.users.map.email", "mail"),
    }

    # Discover flags by LDAP groups
    if _settings.get("ldap.groups.search", None):
        group_type = _settings.get("ldap.groups.type", "groupOfNames")

        # The TOML config might contain either one table or an array of tables
        _AUTH_LDAP_GROUP_SETTINGS = _settings.get("ldap.groups.search")
        if not isinstance(_AUTH_LDAP_GROUP_SETTINGS, list):
            _AUTH_LDAP_GROUP_SETTINGS = [_AUTH_LDAP_GROUP_SETTINGS]

        AUTH_LDAP_GROUP_SEARCH = LDAPSearchUnion(
            *[
                LDAPSearch(
                    entry["base"],
                    ldap.SCOPE_SUBTREE,
                    entry.get("filter", f"(objectClass={group_type})"),
                )
                for entry in _AUTH_LDAP_GROUP_SETTINGS
            ]
        )

        _group_type = _settings.get("ldap.groups.type", "groupOfNames").lower()
        if _group_type == "groupofnames":
            AUTH_LDAP_GROUP_TYPE = NestedGroupOfNamesType()
        elif _group_type == "groupofuniquenames":
            AUTH_LDAP_GROUP_TYPE = NestedGroupOfUniqueNamesType()
        elif _group_type == "posixgroup":
            AUTH_LDAP_GROUP_TYPE = PosixGroupType()

        AUTH_LDAP_USER_FLAGS_BY_GROUP = {}
        for _flag in ["is_active", "is_staff", "is_superuser"]:
            _dn = _settings.get(f"ldap.groups.flags.{_flag}", None)
            if _dn:
                AUTH_LDAP_USER_FLAGS_BY_GROUP[_flag] = _dn

        # Backend admin requires superusers to also be staff members
        if (
            "is_superuser" in AUTH_LDAP_USER_FLAGS_BY_GROUP
            and "is_staff" not in AUTH_LDAP_USER_FLAGS_BY_GROUP
        ):
            AUTH_LDAP_USER_FLAGS_BY_GROUP["is_staff"] = AUTH_LDAP_USER_FLAGS_BY_GROUP[
                "is_superuser"
            ]

CUSTOM_AUTHENTICATION_BACKENDS = []
merge_app_settings("AUTHENTICATION_BACKENDS", CUSTOM_AUTHENTICATION_BACKENDS)

# Add ModelBckend last so all other backends get a chance
# to verify passwords first
AUTHENTICATION_BACKENDS.append("django.contrib.auth.backends.ModelBackend")

# Authentication backend for django-allauth.
AUTHENTICATION_BACKENDS.append("allauth.account.auth_backends.AuthenticationBackend")

# Structure of items: backend, URL name, icon name, button title
ALTERNATIVE_LOGIN_VIEWS = []
merge_app_settings("ALTERNATIVE_LOGIN_VIEWS", ALTERNATIVE_LOGIN_VIEWS, True)

# Internationalization
# https://docs.djangoproject.com/en/2.1/topics/i18n/

LANGUAGES = [
    ("en", _("English")),
    ("de", _("German")),
]
LANGUAGE_CODE = _settings.get("l10n.lang", "en")
TIME_ZONE = _settings.get("l10n.tz", "UTC")
USE_I18N = True
USE_L10N = True
USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.1/howto/static-files/


STATIC_URL = _settings.get("static.url", "/static/")
MEDIA_URL = _settings.get("media.url", "/media/")

LOGIN_REDIRECT_URL = "index"
LOGOUT_REDIRECT_URL = "index"

STATIC_ROOT = _settings.get("static.root", os.path.join(BASE_DIR, "static"))
MEDIA_ROOT = _settings.get("media.root", os.path.join(BASE_DIR, "media"))
NODE_MODULES_ROOT = _settings.get("node_modules.root", os.path.join(BASE_DIR, "node_modules"))

YARN_INSTALLED_APPS = [
    "@fontsource/roboto",
    "jquery",
    "materialize-css",
    "material-design-icons-iconfont",
    "select2",
    "select2-materialize",
    "paper-css",
    "jquery-sortablejs",
    "sortablejs",
]

merge_app_settings("YARN_INSTALLED_APPS", YARN_INSTALLED_APPS, True)

JS_URL = _settings.get("js_assets.url", STATIC_URL)
JS_ROOT = _settings.get("js_assets.root", NODE_MODULES_ROOT + "/node_modules")

SELECT2_CSS = JS_URL + "/select2/dist/css/select2.min.css"
SELECT2_JS = JS_URL + "/select2/dist/js/select2.min.js"
SELECT2_I18N_PATH = JS_URL + "/select2/dist/js/i18n"

ANY_JS = {
    "materialize": {"js_url": JS_URL + "/materialize-css/dist/js/materialize.min.js"},
    "jQuery": {"js_url": JS_URL + "/jquery/dist/jquery.min.js"},
    "material-design-icons": {
        "css_url": JS_URL + "/material-design-icons-iconfont/dist/material-design-icons.css"
    },
    "paper-css": {"css_url": JS_URL + "/paper-css/paper.min.css"},
    "select2-materialize": {
        "css_url": JS_URL + "/select2-materialize/select2-materialize.css",
        "js_url": JS_URL + "/select2-materialize/index.js",
    },
    "sortablejs": {"js_url": JS_URL + "/sortablejs/Sortable.min.js"},
    "jquery-sortablejs": {"js_url": JS_URL + "/jquery-sortablejs/jquery-sortable.js"},
    "Roboto100": {"css_url": JS_URL + "/@fontsource/roboto/100.css"},
    "Roboto300": {"css_url": JS_URL + "/@fontsource/roboto/300.css"},
    "Roboto400": {"css_url": JS_URL + "/@fontsource/roboto/400.css"},
    "Roboto500": {"css_url": JS_URL + "/@fontsource/roboto/500.css"},
    "Roboto700": {"css_url": JS_URL + "/@fontsource/roboto/700.css"},
    "Roboto900": {"css_url": JS_URL + "/@fontsource/roboto/900.css"},
}

merge_app_settings("ANY_JS", ANY_JS, True)

SASS_PROCESSOR_ENABLED = True
SASS_PROCESSOR_AUTO_INCLUDE = False
SASS_PROCESSOR_CUSTOM_FUNCTIONS = {
    "get-colour": "aleksis.core.util.sass_helpers.get_colour",
    "get-preference": "aleksis.core.util.sass_helpers.get_preference",
}
SASS_PROCESSOR_INCLUDE_DIRS = [
    _settings.get("materialize.sass_path", JS_ROOT + "/materialize-css/sass/"),
    STATIC_ROOT + "/materialize-css/sass/",
    STATIC_ROOT,
]

ADMINS = _settings.get("contact.admins", [])
SERVER_EMAIL = _settings.get("contact.from", "root@localhost")
DEFAULT_FROM_EMAIL = _settings.get("contact.from", "root@localhost")
MANAGERS = _settings.get("contact.admins", [])

if _settings.get("mail.server.host", None):
    EMAIL_HOST = _settings.get("mail.server.host")
    EMAIL_USE_TLS = _settings.get("mail.server.tls", False)
    EMAIL_USE_SSL = _settings.get("mail.server.ssl", False)
    if _settings.get("mail.server.port", None):
        EMAIL_PORT = _settings.get("mail.server.port")
    if _settings.get("mail.server.user", None):
        EMAIL_HOST_USER = _settings.get("mail.server.user")
        EMAIL_HOST_PASSWORD = _settings.get("mail.server.password")

TEMPLATED_EMAIL_BACKEND = "templated_email.backends.vanilla_django"
TEMPLATED_EMAIL_AUTO_PLAIN = True

DYNAMIC_PREFERENCES = {
    "REGISTRY_MODULE": "preferences",
}

MAINTENANCE_MODE = _settings.get("maintenance.enabled", None)
MAINTENANCE_MODE_IGNORE_IP_ADDRESSES = _settings.get(
    "maintenance.ignore_ips", _settings.get("maintenance.internal_ips", [])
)
MAINTENANCE_MODE_GET_CLIENT_IP_ADDRESS = "ipware.ip.get_ip"
MAINTENANCE_MODE_IGNORE_SUPERUSER = True
MAINTENANCE_MODE_STATE_FILE_NAME = _settings.get(
    "maintenance.statefile", "maintenance_mode_state.txt"
)
MAINTENANCE_MODE_STATE_BACKEND = "maintenance_mode.backends.DefaultStorageBackend"

DBBACKUP_STORAGE = _settings.get("backup.storage", "django.core.files.storage.FileSystemStorage")
DBBACKUP_STORAGE_OPTIONS = {"location": _settings.get("backup.location", "/var/backups/aleksis")}
DBBACKUP_CLEANUP_KEEP = _settings.get("backup.database.keep", 10)
DBBACKUP_CLEANUP_KEEP_MEDIA = _settings.get("backup.media.keep", 10)
DBBACKUP_GPG_RECIPIENT = _settings.get("backup.gpg_recipient", None)
DBBACKUP_COMPRESS_DB = _settings.get("backup.database.compress", True)
DBBACKUP_ENCRYPT_DB = _settings.get("backup.database.encrypt", DBBACKUP_GPG_RECIPIENT is not None)
DBBACKUP_COMPRESS_MEDIA = _settings.get("backup.media.compress", True)
DBBACKUP_ENCRYPT_MEDIA = _settings.get("backup.media.encrypt", DBBACKUP_GPG_RECIPIENT is not None)
DBBACKUP_CLEANUP_DB = _settings.get("backup.database.clean", True)
DBBACKUP_CLEANUP_MEDIA = _settings.get("backup.media.clean", True)
DBBACKUP_CONNECTOR_MAPPING = {
    "django_prometheus.db.backends.postgresql": "dbbackup.db.postgresql.PgDumpConnector",
}

if _settings.get("backup.storage.type", "").lower() == "s3":
    DBBACKUP_STORAGE = "storages.backends.s3boto3.S3Boto3Storage"

    DBBACKUP_STORAGE_OPTIONS = {
        key: value for (key, value) in _settings.get("backup.storage.s3").items()
    }

IMPERSONATE = {"USE_HTTP_REFERER": True, "REQUIRE_SUPERUSER": True, "ALLOW_SUPERUSER": True}

DJANGO_TABLES2_TEMPLATE = "django_tables2/materialize.html"

ANONYMIZE_ENABLED = _settings.get("maintenance.anonymisable", True)

LOGIN_URL = "two_factor:login"

if _settings.get("2fa.call.enabled", False):
    if "two_factor.middleware.threadlocals.ThreadLocals" not in MIDDLEWARE:
        MIDDLEWARE.insert(
            MIDDLEWARE.index("django_otp.middleware.OTPMiddleware") + 1,
            "two_factor.middleware.threadlocals.ThreadLocals",
        )
    TWO_FACTOR_CALL_GATEWAY = "two_factor.gateways.twilio.gateway.Twilio"

if _settings.get("2fa.sms.enabled", False):
    if "two_factor.middleware.threadlocals.ThreadLocals" not in MIDDLEWARE:
        MIDDLEWARE.insert(
            MIDDLEWARE.index("django_otp.middleware.OTPMiddleware") + 1,
            "two_factor.middleware.threadlocals.ThreadLocals",
        )
    TWO_FACTOR_SMS_GATEWAY = "two_factor.gateways.twilio.gateway.Twilio"

if _settings.get("twilio.sid", None):
    TWILIO_SID = _settings.get("twilio.sid")
    TWILIO_TOKEN = _settings.get("twilio.token")
    TWILIO_CALLER_ID = _settings.get("twilio.callerid")

CELERY_BROKER_URL = _settings.get("celery.broker", REDIS_URL)
CELERY_RESULT_BACKEND = "django-db"
CELERY_CACHE_BACKEND = "django-cache"
CELERY_BEAT_SCHEDULER = "django_celery_beat.schedulers:DatabaseScheduler"

if _settings.get("celery.email", False):
    EMAIL_BACKEND = "djcelery_email.backends.CeleryEmailBackend"

if _settings.get("dev.uwsgi.celery", DEBUG):
    concurrency = _settings.get("celery.uwsgi.concurrency", 2)
    UWSGI.setdefault("attach-daemon", [])
    UWSGI["attach-daemon"].append(f"celery -A aleksis.core worker --concurrency={concurrency}")
    UWSGI["attach-daemon"].append("celery -A aleksis.core beat")

DEFAULT_FAVICON_PATHS = {
    "pwa_icon": os.path.join(STATIC_ROOT, "img/aleksis-icon.png"),
    "favicon": os.path.join(STATIC_ROOT, "img/aleksis-icon.png"),
}
PWA_ICONS_CONFIG = {
    "android": [192, 512],
    "apple": [76, 114, 152, 180],
    "apple_splash": [192],
    "microsoft": [144],
}

SERVICE_WORKER_PATH = os.path.join(STATIC_ROOT, "js", "serviceworker.js")

SITE_ID = 1

CKEDITOR_CONFIGS = {
    "default": {
        "toolbar_Basic": [["Source", "-", "Bold", "Italic"]],
        "toolbar_Full": [
            {
                "name": "document",
                "items": ["Source", "-", "Save", "NewPage", "Preview", "Print", "-", "Templates"],
            },
            {
                "name": "clipboard",
                "items": [
                    "Cut",
                    "Copy",
                    "Paste",
                    "PasteText",
                    "PasteFromWord",
                    "-",
                    "Undo",
                    "Redo",
                ],
            },
            {"name": "editing", "items": ["Find", "Replace", "-", "SelectAll"]},
            {
                "name": "insert",
                "items": [
                    "Image",
                    "Table",
                    "HorizontalRule",
                    "Smiley",
                    "SpecialChar",
                    "PageBreak",
                    "Iframe",
                ],
            },
            "/",
            {
                "name": "basicstyles",
                "items": [
                    "Bold",
                    "Italic",
                    "Underline",
                    "Strike",
                    "Subscript",
                    "Superscript",
                    "-",
                    "RemoveFormat",
                ],
            },
            {
                "name": "paragraph",
                "items": [
                    "NumberedList",
                    "BulletedList",
                    "-",
                    "Outdent",
                    "Indent",
                    "-",
                    "Blockquote",
                    "CreateDiv",
                    "-",
                    "JustifyLeft",
                    "JustifyCenter",
                    "JustifyRight",
                    "JustifyBlock",
                    "-",
                    "BidiLtr",
                    "BidiRtl",
                    "Language",
                ],
            },
            {"name": "links", "items": ["Link", "Unlink", "Anchor"]},
            "/",
            {"name": "styles", "items": ["Styles", "Format", "Font", "FontSize"]},
            {"name": "colors", "items": ["TextColor", "BGColor"]},
            {"name": "tools", "items": ["Maximize", "ShowBlocks"]},
            {"name": "about", "items": ["About"]},
            {"name": "customtools", "items": ["Preview", "Maximize",]},
        ],
        "toolbar": "Full",
        "tabSpaces": 4,
        "extraPlugins": ",".join(
            [
                "uploadimage",
                "div",
                "autolink",
                "autoembed",
                "embedsemantic",
                "autogrow",
                # 'devtools',
                "widget",
                "lineutils",
                "clipboard",
                "dialog",
                "dialogui",
                "elementspath",
            ]
        ),
    }
}

# Upload path for CKEditor. Relative to MEDIA_ROOT.
CKEDITOR_UPLOAD_PATH = "ckeditor_uploads/"

# Which HTML tags are allowed
BLEACH_ALLOWED_TAGS = ["p", "b", "i", "u", "em", "strong", "a", "div"]

# Which HTML attributes are allowed
BLEACH_ALLOWED_ATTRIBUTES = ["href", "title", "style"]

# Which CSS properties are allowed in 'style' attributes (assuming
# style is an allowed attribute)
BLEACH_ALLOWED_STYLES = ["font-family", "font-weight", "text-decoration", "font-variant"]

# Strip unknown tags if True, replace with HTML escaped characters if
# False
BLEACH_STRIP_TAGS = True

# Strip comments, or leave them in.
BLEACH_STRIP_COMMENTS = True

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {"console": {"class": "logging.StreamHandler", "formatter": "verbose"},},
    "formatters": {"verbose": {"format": "%(levelname)s %(asctime)s %(module)s: %(message)s"}},
    "root": {"handlers": ["console"], "level": _settings.get("logging.level", "WARNING"),},
}

# Rules and permissions

GUARDIAN_RAISE_403 = True
ANONYMOUS_USER_NAME = None

SILENCED_SYSTEM_CHECKS.append("guardian.W001")

# Append authentication backends
AUTHENTICATION_BACKENDS.append("rules.permissions.ObjectPermissionBackend")

HAYSTACK_CONNECTIONS = {
    "default": {"ENGINE": "haystack_redis.RedisEngine", "PATH": REDIS_URL,},
}

HAYSTACK_SIGNAL_PROCESSOR = "celery_haystack.signals.CelerySignalProcessor"
CELERY_HAYSTACK_IGNORE_RESULT = True

HAYSTACK_SEARCH_RESULTS_PER_PAGE = 10

DJANGO_EASY_AUDIT_WATCH_REQUEST_EVENTS = False

HEALTH_CHECK = {
    "DISK_USAGE_MAX": _settings.get("health.disk_usage_max_percent", 90),
    "MEMORY_MIN": _settings.get("health.memory_min_mb", 500),
}

DBBACKUP_CHECK_SECONDS = _settings.get("backup.database.check_seconds", 7200)
MEDIABACKUP_CHECK_SECONDS = _settings.get("backup.media.check_seconds", 7200)

PROMETHEUS_EXPORT_MIGRATIONS = False
PROMETHEUS_METRICS_EXPORT_PORT = _settings.get("prometheus.metrics.port", None)
PROMETHEUS_METRICS_EXPORT_ADDRESS = _settings.get("prometheus.metrucs.address", None)

SECURE_PROXY_SSL_HEADER = ("REQUEST_SCHEME", "https")

if _settings.get("storage.type", "").lower() == "s3":
    INSTALLED_APPS.append("storages")

    DEFAULT_FILE_STORAGE = "storages.backends.s3boto3.S3Boto3Storage"

    if _settings.get("storage.s3.static.enabled", False):
        STATICFILES_STORAGE = "storages.backends.s3boto3.S3StaticStorage"
        AWS_STORAGE_BUCKET_NAME_STATIC = _settings.get("storage.s3.static.bucket_name", "")
        AWS_S3_MAX_AGE_SECONDS_CACHED_STATIC = _settings.get(
            "storage.s3.static.max_age_seconds", 24 * 60 * 60
        )

    AWS_REGION = _settings.get("storage.s3.region_name", "")
    AWS_ACCESS_KEY_ID = _settings.get("storage.s3.access_key", "")
    AWS_SECRET_ACCESS_KEY = _settings.get("storage.s3.secret_key", "")
    AWS_SESSION_TOKEN = _settings.get("storage.s3.session_token", "")
    AWS_STORAGE_BUCKET_NAME = _settings.get("storage.s3.bucket_name", "")
    AWS_LOCATION = _settings.get("storage.s3.location", "")
    AWS_S3_ADDRESSING_STYLE = _settings.get("storage.s3.addressing_style", "auto")
    AWS_S3_ENDPOINT_URL = _settings.get("storage.s3.endpoint_url", "")
    AWS_S3_KEY_PREFIX = _settings.get("storage.s3.key_prefix", "")
    AWS_S3_BUCKET_AUTH = _settings.get("storage.s3.bucket_auth", True)
    AWS_S3_MAX_AGE_SECONDS = _settings.get("storage.s3.max_age_seconds", 24 * 60 * 60)
    AWS_S3_PUBLIC_URL = _settings.get("storage.s3.public_url", "")
    AWS_S3_REDUCED_REDUNDANCY = _settings.get("storage.s3.reduced_redundancy", False)
    AWS_S3_CONTENT_DISPOSITION = _settings.get("storage.s3.content_disposition", "")
    AWS_S3_CONTENT_LANGUAGE = _settings.get("storage.s3.content_language", "")
    AWS_S3_METADATA = _settings.get("storage.s3.metadata", {})
    AWS_S3_ENCRYPT_KEY = _settings.get("storage.s3.encrypt_key", False)
    AWS_S3_KMS_ENCRYPTION_KEY_ID = _settings.get("storage.s3.kms_encryption_key_id", "")
    AWS_S3_GZIP = _settings.get("storage.s3.gzip", True)
    AWS_S3_SIGNATURE_VERSION = _settings.get("storage.s3.signature_version", None)
    AWS_S3_FILE_OVERWRITE = _settings.get("storage.s3.file_overwrite", False)
    AWS_S3_VERIFY = _settings.get("storage.s3.verify", True)
    AWS_S3_USE_SSL = _settings.get("storage.s3.use_ssl", True)
else:
    DEFAULT_FILE_STORAGE = "titofisto.TitofistoStorage"
    TITOFISTO_TIMEOUT = 10 * 60

SASS_PROCESSOR_STORAGE = DEFAULT_FILE_STORAGE

# Add django-cleanup after all apps to ensure that it gets all signals as last app
INSTALLED_APPS.append("django_cleanup.apps.CleanupConfig")



import pytest

from aleksis.core.models import Notification, Person

pytestmark = pytest.mark.django_db


def test_email_notification(mailoutbox):
    email = "doe@example.com"
    recipient = Person.objects.create(first_name="Jane", last_name="Doe", email=email)

    sender = "Foo"
    title = "There is happened something."
    description = "Here you get some more information."
    link = "https://aleksis.org/"

    notification = Notification(
        sender=sender, recipient=recipient, title=title, description=description, link=link
    )
    notification.save()

    assert notification.sent

    assert len(mailoutbox) == 1

    mail = mailoutbox[0]

    assert email in mail.to
    assert title in mail.body
    assert description in mail.body
    assert link in mail.body
    assert sender in mail.body
    assert recipient.addressing_name in mail.subject
    assert recipient.addressing_name in mail.body
